import re

from typing import Dict, List

import arrow
from datetime import datetime

from canvas_workflow_kit.timeframe import Timeframe


def get_by_path(dic, keys):
    """Access a nested object in root by item sequence."""
    for key in keys:
        dic = dic.get(key, None)
        if dic is None:
            break
    return dic


URL_CPT = 'http://www.ama-assn.org/go/cpt'
URL_HCPCS = 'https://coder.aapc.com/hcpcs-codes'
URL_CVX = 'http://hl7.org/fhir/sid/cvx'
URL_LOINC = 'http://loinc.org'
URL_SNOMED = 'http://snomed.info/sct'
URL_FDB = 'http://www.fdbhealth.com/'
URL_RXNORM = 'http://www.nlm.nih.gov/research/umls/rxnorm'
URL_ICD10 = 'ICD-10'
URL_NUCC = 'http://www.nucc.org/'
URL_CANVAS = 'CANVAS'
URL_INTERNAL = 'INTERNAL'
URL_NDC = 'http://hl7.org/fhir/sid/ndc'

CODE_SYSTEM_MAPPING = {
    'cpt': URL_CPT,
    'hcpcs': URL_HCPCS,
    'cvx': URL_CVX,
    'loinc': URL_LOINC,
    'snomedct': URL_SNOMED,
    'fdb': URL_FDB,
    'rxnorm': URL_RXNORM,
    'icd10cm': URL_ICD10,
    'icd10pcs': URL_ICD10,
    'nucc': URL_NUCC,
    'canvas': URL_CANVAS,
    'internal': URL_INTERNAL,
    'ndc': URL_NDC
}

SYSTEM_CODE_MAPPING = {
    URL_CPT: ['cpt', 'hcpcs'],
    URL_CVX: ['cvx'],
    URL_LOINC: ['loinc'],
    URL_SNOMED: ['snomedct'],
    URL_FDB: ['fdb'],
    URL_RXNORM: ['rxnorm'],
    URL_ICD10: ['icd10cm', 'icd10pcs'],
    URL_NUCC: ['nucc'],
    URL_CANVAS: ['canvas'],
    URL_INTERNAL: ['internal'],
    URL_NDC: ['ndc']
}


class PatientRecordSet:
    """
    Base class representing sorted sets of records (conditions, immunizations,
    etc) for a patient.  You can iterate through them and get at the
    record data:

        for condition in patient.conditions:
            print(f'Sorry, you have {condition['name']}')

    To filter on objects use filters:
        patient.referrals.filter(order=None)  # find referrals where `"order": null`
        patient.referrals.filter(timestamp__lt=arrow.now())

    Filters can use any of the following appended to the property name:
        exact, iexact, contains, icontains,
        gt, gte, lt, lte,
        startswith, endswith, istartswith, iendswith

    If not specified, exact will be used by default.

    To filter on a nested objects, use double underscore.
        patient.referral.filter(report__status='NEW')

    You can find records by value set:

        patient.conditions.find(Diabetes)

    Filter by date in various ways:

        patient.conditions.within(timeframe)
        patient.conditions.before(now)
        patient.conditions.after(yesterday)

    Methods are chainable:

        patient.conditions.find(Diabetes).within(timeframe)

    Except these, which pick the last record and extract a value from it:

        patient.conditions.find(Diabetes).within(timeframe).last()
        patient.lab_reports.find(Hba1CLaboratoryTest).within(timeframe).last_value()

    """
    API_KEY_FIELD = 'patient__key'
    API_UPDATE_FIELD = None
    VALID_SYSTEMS = None
    DATE_FIELD = 'date'

    operators = {
        'exact': lambda x, y: x == y,
        'iexact': lambda x, y: x.lower() == y.lower(),
        'contains': lambda x, y: x in y,
        'icontains': lambda x, y: x.lower() in y.lower(),
        'gt': lambda x, y: x > y,
        'gte': lambda x, y: x >= y,
        'lt': lambda x, y: x < y,
        'lte': lambda x, y: x <= y,
        'startswith': lambda x, y: False if not isinstance(x, str) else x.startswith(y),
        'endswith': lambda x, y: False if not isinstance(x, str) else x.endswith(y),
        'istartswith': lambda x, y: False if not isinstance(x, str) else x.lower().startswith(y.lower()),
        'iendswith': lambda x, y: False if not isinstance(x, str) else x.lower().endswith(y.lower()),
    }

    def __init__(self, records):
        self.records = records

    def __str__(self):
        return f'<{self.__class__.__name__}: {len(self.records)} records>'

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return self.records.__iter__()

    def __getitem__(self, index):
        return self.records[index]

    def __bool__(self):
        return True if self.records else False

    def __len__(self):
        return len(self.records)

    def __add__(self, other):
        if self.__class__ != other.__class__:
            raise Exception(f'Illegal attempt to add a {self.__class__} and a {other.__class__}.')

        return self.__class__(self.records + other.records)

    @classmethod
    def new_from(cls, records: List[Dict]):
        return cls(records)

    @staticmethod
    def item_to_codes(item) -> List:
        """
        Internal method used to find the codes in a given record
        object, should be overriden by sub-classes as needed.
        """
        if 'code' in item:
            return item['code']['coding']

        if 'coding' in item:
            return item['coding']

        if 'substance' in item:
            return item['substance']['coding']

        return []

    def validated(self):
        """
        Filter records that have been committed, not deleted or enter by error.
        Returns a PatientRecordSet object of the same sub-class, enabling chaining of calls.
        """
        items = []
        for item in self.records:
            if item.get('deleted'):
                continue

            if item.get('entered_in_error') or item.get('enteredInError'):
                continue

            if not item.get('committer'):
                continue

            items.append(item)

        return self.__class__(items)

    def _enumerate_system_codes(self, item):
        """
        Returns systems, and code from a coding.
        Eg: ['icd10cm', 'icd10pcs'], E119
        """
        # not all patient record data has codes, ex: referrals
        if not self.VALID_SYSTEMS:
            raise Exception('Invalid call to find() on non-coded patient data.')

        for coding in self.item_to_codes(item):
            coding_systems = SYSTEM_CODE_MAPPING.get(coding['system'], [])
            for coding_system in coding_systems:
                yield coding_systems, coding.get('code')

    def _enumerate_system_list_codes(self, value_set):
        # not all patient record data has codes, ex: referrals
        if not self.VALID_SYSTEMS:
            raise Exception('Invalid call to find() on non-coded patient data.')

        values = value_set.values
        systems = [key for key in values.keys() if key in self.VALID_SYSTEMS]

        for item in self.records:
            for coding in self.item_to_codes(item):
                coding_systems = [
                    sys for sys in SYSTEM_CODE_MAPPING.get(coding['system'], []) if sys in systems
                ]

                for coding_system in coding_systems:
                    yield item, coding, values[coding_system]

    def _parse_shorthand_filter(self, key):
        tokens = key.split("__")

        operator_name = 'exact'
        operator_method = self.operators['exact']
        if tokens[-1] in self.operators:
            operator_name = tokens.pop()
            operator_method = self.operators.get(operator_name)

        def left_side(item):
            left_side_value = get_by_path(item, tokens)
            return left_side_value

        def filter_fn(x, y):
            left_val = left_side(x)

            # Modify types so they cna be compared
            # TODO: do date conversion in the record object itself
            if isinstance(y, (arrow.Arrow, datetime)) and isinstance(left_val, str):
                left_val = arrow.get(left_val)

            elif isinstance(y, (arrow.Arrow, datetime)) and left_val is None:
                return False

            return operator_method(left_val, y)
        return filter_fn

    def filter(self, **kwargs):
        """
        Filter a queryset.
        To filter on a nested objects, use double underscore.
          Ex: .filter(report__status='NEW')
        Django operators can also be used.
          Ex: .filter(report__status__istartswith='ne', report__status__icontains='w')
        """
        queryset = self.__class__(self.records)

        for kwarg_key, value in kwargs.items():
            filter_fn = self._parse_shorthand_filter(kwarg_key)
            queryset.records = [
                item for item in queryset if filter_fn(item, value)
            ]
        return queryset

    def exclude(self, **kwargs):
        """
        Exclude values in a queryset using the same method as filter()
        """

        queryset = self.__class__(self.records)

        for kwarg_key, value in kwargs.items():
            filter_fn = self._parse_shorthand_filter(kwarg_key)
            queryset.records = [
                item for item in queryset if not filter_fn(item, value)
            ]
        return queryset

    def find_code(self, *args, **kwargs):
        """
        In development.
        Not for external use.

        Returns matching records by code system and code, for when a ValueSet is not available.
        Ex: patient.conditions.find_code(icd10cm="E119")
        """
        queryset = self.__class__(self.records)

        def filter_fn(record):
            for codings, code in self._enumerate_system_codes(record):
                for key, value in kwargs.items():
                    if key.lower() in codings and code.upper() == value.upper():
                        return True
            return False

        queryset.records = [
            record for record in queryset if filter_fn(record)
        ]

        return queryset

    def find(self, value_set):
        """
        Find records that match a given value_set based on exact code. Returns a
        PatientRecordSet object of the same sub-class, enabling chaining
        of calls.
        """
        items = []
        for item, coding, list_codes in self._enumerate_system_list_codes(value_set):
            if coding['code'] in list_codes:
                items.append(item)

        return self.__class__(items)

    def find_class(self, value_set):
        """
        Find records that match a given value_set based on regular expression. Returns a
        PatientRecordSet object of the same sub-class, enabling chaining
        of calls.
        """
        items = []
        for item, coding, list_codes in self._enumerate_system_list_codes(value_set):
            for code_re in list_codes:
                if re.search(code_re, coding['code']):
                    items.append(item)
                    break

        return self.__class__(items)


class PatientEventRecordSet(PatientRecordSet):

    def __init__(self, records):
        super().__init__(records)
        # filter out records with no date attached - they're useless
        # for protocols since everything is constrained by timeframe
        records = [r for r in records if r[self.DATE_FIELD] is not None]
        self.records = sorted(records, key=lambda rec: rec[self.DATE_FIELD])

    def filter_function(self, filter_fn):
        """
        Filter records by date, pass in a function and it will be
        applied to each date.  Returns a PatientRecordSet object of the
        same sub-class, enabling chaining of calls.
        """

        

        return self.__class__([
            item for item in self.records
            if item[self.DATE_FIELD] and filter_fn(arrow.get(item[self.DATE_FIELD]))
        ])

    def within(self, timeframe: Timeframe):
        """
        Find records within a timeframe.  Returns a PatientRecordSet
        object of the same sub-class, enabling chaining of calls.
        """

        def filter_fn(onset) -> bool:
            return timeframe.start <= onset <= timeframe.end

        return self.filter_function(filter_fn)

    def before(self, end: arrow):
        """
        Find records before an arrow datetime.  Returns a
        PatientRecordSet object of the same sub-class, enabling chaining
        of calls.
        """
        return self.filter_function(lambda onset: onset <= end)

    def after(self, start: arrow):
        """
        Find records after an arrow datetime.  Returns a
        PatientRecordSet object of the same sub-class, enabling chaining
        of calls.
        """
        return self.filter_function(lambda onset: onset >= start)

    # TODO_REPORTING the "Measurement Period" is defined as the period from
    # Jan 1st to Dec 31st of a specified year
    def within_one_year(self, timeframe: Timeframe):
        """
        Find records within the year before the timeframe end. Returns a
        PatientRecordSet object of the same sub-class, enabling chaining
        of calls.
        """
        return self.within(Timeframe(start=timeframe.end.shift(years=-1), end=timeframe.end))

    # TODO_REPORTING the "Measurement Period" is defined as the period from
    # Jan 1st to Dec 31st of a specified year
    def within_two_years(self, timeframe: Timeframe):
        """
        Find records within the 2 years before the timeframe end. Returns a
        PatientRecordSet object of the same sub-class, enabling chaining
        of calls.
        """
        return self.within(Timeframe(start=timeframe.end.shift(years=-2), end=timeframe.end))

    def first(self):
        """
        Returns the first (less recent) record.
        """
        if self.records:
            return self.records[0]
        return None

    def last(self):
        """
        Returns the last (most recent) record.
        """
        if self.records:
            return self.records[-1]
        return None

    def last_value(self, cast=None):
        """
        Returns the last (most recent) record.
        """
        record = self.last()

        if record is None:
            return None

        if cast:
            try:
                return cast(record['value'])
            except (ValueError, TypeError):
                return None
        else:
            return record['value']


class PatientPeriodRecordSet(PatientRecordSet):

    def __init__(self, records):
        super().__init__(records)
        self.records = [r for r in records if len(r['periods']) > 0]

    def intersects(self, timeframe: Timeframe, still_active: bool):
        """
        Search all records with a period overlapping with the provided period
        Returns a PatientRecordSet object of the same sub-class, enabling chaining of calls.

        still_active is used in regard of the end date of the record (period['to']) when
        it is known (i.e. not None)
        The record is validating the criteria
         - when still_active==False, this means that the record is active during the period
         - when still_active==True , this means that the record is still active after the
         end of the period
        """
        result = []
        for item in self.records:
            for period in item['periods']:
                if (arrow.get(period['from']) <= timeframe.end and
                        (period['to'] is None or
                         ((not still_active and
                           timeframe.start <= arrow.get(period['to'])) or
                          (still_active and
                           timeframe.end <= arrow.get(period['to']))))):  # yapf: disable
                    result.append(item)
                    break

        return self.__class__(result)

    def starts_before(self, on_date: arrow):
        """
        Search all records with a still active period starting before the end of
        the provided period
        Returns a PatientRecordSet object of the same sub-class, enabling chaining of calls.
        """
        result = []
        for item in self.records:
            for period in item['periods']:
                if arrow.get(period['from']) <= on_date and period['to'] is None:
                    result.append(item)

        return self.__class__(result)


class BillingLineItemRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS = ['cpt', 'hcpcs', 'snomedct']
    DATE_FIELD = 'datetimeOfService'
    PATIENT_FIELD = 'billing_line_items'
    API_UPDATE_FIELD = 'billingLineItems'

    @staticmethod
    def item_to_codes(item) -> List:
        return [{'code': item['cpt'], 'system': URL_CPT}]


class ConditionRecordSet(PatientPeriodRecordSet):
    VALID_SYSTEMS = ['icd10cm', 'icd10pcs', 'snomedct']
    PATIENT_FIELD = 'conditions'
    API_UPDATE_FIELD = 'conditions'

    @staticmethod
    def item_to_codes(item) -> List:
        if item['clinicalStatus'] in ['active', 'resolved']:
            return item['coding']
        return []


class ImmunizationRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS = ['cvx']
    DATE_FIELD = 'dateOrdered'
    PATIENT_FIELD = 'immunizations'
    API_UPDATE_FIELD = 'immunizations'

    @staticmethod
    def item_to_codes(item) -> List:
        return item['coding']


class VitalSignRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS = ['loinc']
    DATE_FIELD = 'dateRecorded'
    PATIENT_FIELD = 'vital_signs'
    API_KEY_FIELD = 'reading__patient__key'
    API_UPDATE_FIELD = 'vitalSigns'

    @staticmethod
    def item_to_codes(item) -> List:
        return [{'code': item['loincNum'], 'system': URL_LOINC}]


class LabReportRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS = ['loinc']
    DATE_FIELD = 'originalDate'
    PATIENT_FIELD = 'lab_reports'
    API_UPDATE_FIELD = 'labReports'

    @staticmethod
    def item_to_codes(item) -> List:
        codes = []
        for code in item['loincCodes']:
            codes.append({'system': URL_LOINC, **code})
        return codes

    @classmethod
    def new_from(cls, data: Dict):
        # lab reports come in a slightly weird structure - the records we
        # care about are the lab values but the date we need to filter on
        # is in the outer lab-report structure.  To use the
        # PatientRecordSet code we need to collapse them so we have a date
        # stamp and a set of codes in a single record hash.
        combined = []
        for report in data:
            for value in report['values']:
                combined.append({'originalDate': report['originalDate'], **value})

        return cls(combined)


class InstructionRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS = ['icd10cm', 'icd10pcs', 'snomedct', 'cpt', 'hcpcs']
    DATE_FIELD = 'noteTimestamp'
    PATIENT_FIELD = 'instructions'
    API_UPDATE_FIELD = 'instructions'


class MedicationRecordSet(PatientPeriodRecordSet):
    VALID_SYSTEMS = [
        'rxnorm',
        'fdb',
    ]
    PATIENT_FIELD = 'medications'
    API_UPDATE_FIELD = 'medications'

    @staticmethod
    def item_to_codes(item) -> List:
        return item['coding']


class InterviewRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS = ['loinc', 'snomedct', 'rxnorm', 'canvas', 'internal']
    DATE_FIELD = 'noteTimestamp'
    PATIENT_FIELD = 'interviews'
    API_UPDATE_FIELD = 'interviews'

    @staticmethod
    def item_to_codes(item) -> List:
        codes = []

        for result in item['results']:
            if result['code'] != '' and result['codeSystem'] != '':
                codes.append({
                    'code': result['code'],
                    'system': result['codeSystem'],
                    'question_id': None,
                })

        for coding in item.get('questionnaires', []):
            codes.append({
                'code': coding['code'],
                'system': coding['codeSystem'],
                'question_id': None,
            })
        for coding in item.get('questions', []):
            codes.append({
                'code': coding['code'],
                'system': coding['codeSystem'],
                'question_id': coding['questionResponseId'],
            })
        for coding in item.get('responses', []):
            codes.append({
                'code': coding['code'],
                'system': coding['codeSystem'],
                'question_id': coding['questionResponseId'],
            })
        # interviews.find(...)
        return codes

    def find_question_response(self, value_set_question, value_set_response):
        items = []
        questions = []

        for _, coding, list_codes in self._enumerate_system_list_codes(value_set_question):
            if coding['code'] in list_codes:
                questions.append(coding['question_id'])

        for item, coding, list_codes in self._enumerate_system_list_codes(value_set_response):
            if coding['code'] in list_codes and coding['question_id'] in questions:
                items.append(item)

        return self.__class__(items)


class ReferralReportRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS: List[str] = [
        'snomedct',
        'loinc',
        'cpt',
    ]
    DATE_FIELD = 'originalDate'
    PATIENT_FIELD = 'referral_reports'
    API_UPDATE_FIELD = 'referralReports'

    @staticmethod
    def item_to_codes(item) -> List:
        codes = []

        for result in item['codings']:
            codes.append({
                'code': result['code'],
                'system': result['system'],
            })
        return codes


class ReferralRecordSet(PatientEventRecordSet):
    # TO BE DEVELOPED
    VALID_SYSTEMS: List[str] = [
        'snomedct',
        'loinc',
        'cpt',
    ]
    DATE_FIELD = 'timestamp'
    PATIENT_FIELD = 'referrals'
    API_UPDATE_FIELD = 'referrals'


class InpatientStayRecordSet(PatientEventRecordSet):
    # TO BE DEVELOPED
    VALID_SYSTEMS: List[str] = [
        'snomedct',
        'loinc',
        'cpt',
    ]
    DATE_FIELD = 'noteTimestamp'
    PATIENT_FIELD = 'inpatient_stays'
    API_UPDATE_FIELD = 'inpatientStay'


class ImagingReportRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS: List[str] = [
        'snomedct',
        'loinc',
        'cpt',
    ]
    DATE_FIELD = 'originalDate'
    PATIENT_FIELD = 'imaging_reports'
    API_UPDATE_FIELD = 'imagingReports'

    @staticmethod
    def item_to_codes(item) -> List:
        codes = []

        for result in item['codings']:
            codes.append({
                'code': result['code'],
                'system': result['system'],
            })
        return codes


class PrescriptionRecordSet(PatientRecordSet):
    VALID_SYSTEMS = [
        'rxnorm',
        'fdb',
        'ndc'
    ]
    PATIENT_FIELD = 'prescriptions'
    API_UPDATE_FIELD = 'prescriptions'

    @staticmethod
    def item_to_codes(item) -> List:
        return item['coding']


class ProcedureRecordSet(PatientRecordSet):
    VALID_SYSTEMS = [
        'cpt',
        'snomedct'
    ]
    PATIENT_FIELD = 'procedures'
    API_UPDATE_FIELD = 'procedures'

    @staticmethod
    def item_to_codes(item) -> List:
        return item['coding']


class ProtocolOverrideRecordSet(PatientRecordSet):
    VALID_SYSTEMS: List[str] = ['snomedct']
    PATIENT_FIELD = 'protocol_overrides'
    API_UPDATE_FIELD = 'protocolOverrides'

    def defined_for(self, protocol_keys: List[str]):
        for item in self.records:
            if item['protocolKey'] in protocol_keys:
                return item
        return None

    @property
    def all_switched_off(self) -> bool:
        for item in self.records:
            if item['protocolKey'] == '*':
                return True
        return False

    def is_snoozed(self, protocol_keys: List[str]) -> bool:

        def protocol_snoozed(item):
            if (item['protocolKey'] in protocol_keys and item['snooze'] and
                    arrow.now() < arrow.get(
                        item['snooze']['reference']).shift(days=item['snooze']['snoozedDays'])):
                return True

        return any(protocol_snoozed(override) for override in self.records)

    @staticmethod
    def item_to_codes(item) -> List:
        snooze = item.get("snooze", {})
        code = snooze.get("reasonCode")
        system = snooze.get("reasonCodeSystem")
        if code:
            return [{
                'code': code,
                'system': system,
            }]
        return []


class ReasonForVisitRecordSet(PatientEventRecordSet):
    # TO BE DEVELOPED
    VALID_SYSTEMS: List[str] = [
        'snomedct',
        'loinc',
        'cpt',
    ]
    DATE_FIELD = 'datetimeOfService'
    PATIENT_FIELD = 'reason_for_visits'
    API_UPDATE_FIELD = 'reasonForVisits'


class UpcomingAppointmentRecordSet(PatientRecordSet):
    VALID_SYSTEMS: List[str] = ['canvas', 'internal']
    PATIENT_FIELD = 'upcoming_appointments'
    API_UPDATE_FIELD = 'upcomingAppointments'
    DATE_FIELD = 'startTime'


class UpcomingAppointmentNoteRecordSet(PatientRecordSet):
    VALID_SYSTEMS: List[str] = ['canvas', 'internal']
    PATIENT_FIELD = 'upcoming_appointment_notes'
    API_UPDATE_FIELD = 'upcomingAppointmentNotes'


class MessageRecordSet(PatientRecordSet):
    VALID_SYSTEMS: List[str] = ['canvas', 'internal']
    PATIENT_FIELD = 'messages'
    API_UPDATE_FIELD = 'messages'


class AllergyIntoleranceRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS: List[str] = ['fdb']
    PATIENT_FIELD = 'allergy_intolerances'
    API_UPDATE_FIELD = 'allergyIntolerances'
    DATE_FIELD = 'onsetDate'


class AdministrativeDocumentRecordSet(PatientEventRecordSet):
    VALID_SYSTEMS: List[str] = ['canvas', 'internal']
    PATIENT_FIELD = 'administrative_documents'
    API_UPDATE_FIELD = 'administrativeDocuments'
    DATE_FIELD = 'originalDate'



from ..value_set import ValueSet


class FrailtySymptom(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for symptoms of frailty.

    **Data Element Scope:** This value set may use a model element related to Symptom.

    **Inclusion Criteria:** Includes concepts that represent a symptom of frailty.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS134v10, CMS165v10, CMS131v10, CMS122v10, CMS125v10, CMS130v10
    """

    VALUE_SET_NAME = 'Frailty Symptom'
    OID = '2.16.840.1.113883.3.464.1003.113.12.1075'
    DEFINITION_VERSION = '20190315'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    ICD10CM = {
        'R26.0',  # Ataxic gait
        'R26.1',  # Paralytic gait
        'R26.2',  # Difficulty in walking, not elsewhere classified
        'R26.89',  # Other abnormalities of gait and mobility
        'R26.9',  # Unspecified abnormalities of gait and mobility
        'R41.81',  # Age-related cognitive decline
        'R53.1',  # Weakness
        'R53.81',  # Other malaise
        'R53.83',  # Other fatigue
        'R54',  # Age-related physical debility
        'R62.7',  # Adult failure to thrive
        'R63.4',  # Abnormal weight loss
        'R63.6',  # Underweight
        'R64',  # Cachexia
    }
    SNOMEDCT = {
        '4468000',  # Oppenheim's gait (finding)
        '8510008',  # Reduced mobility (finding)
        '13791008',  # Asthenia (finding)
        '18726006',  # Senile asthenia (finding)
        '20940004',  # Spinal hemiparesis (finding)
        '22090007',  # Scissoring gait (finding)
        '22325002',  # Abnormal gait (finding)
        '22631008',  # Unsteady when walking (finding)
        '23042008',  # Spinal paraparesis (finding)
        '25136009',  # Ataxic gait (finding)
        '26544005',  # Muscle weakness (finding)
        '41786007',  # Neurological muscle weakness (finding)
        '43005009',  # Shuffling gait (finding)
        '50314001',  # Partial bilateral paresis (finding)
        '67141003',  # Antalgic gait (finding)
        '78119002',  # Complete bilateral paresis (finding)
        '78691002',  # Staggering gait (finding)
        '84229001',  # Fatigue (finding)
        '89362005',  # Weight loss (finding)
        '102492002',  # Failure to maintain weight (finding)
        '102568007',  # Paresis of lower extremity (finding)
        '102891000',  # Age-related cognitive decline (finding)
        '105501005',  # Dependence on enabling machine or device (finding)
        '105503008',  # Dependence on wheelchair (finding)
        '105504002',  # Dependence on walking stick (finding)
        '126013009',  # Subjective muscle weakness (finding)
        '127378008',  # Bilateral paresis (finding)
        '135834002',  # Pseudoparalysis (finding)
        '160681005',  # Mobile outside with aid (finding)
        '160683008',  # Needs walking aid in home (finding)
        '160684002',  # Confined to chair (finding)
        '160685001',  # Bed-ridden (finding)
        '160692006',  # Mobility very poor (finding)
        '160693001',  # Mobility poor (finding)
        '160734000',  # Lives in a nursing home (finding)
        '160737007',  # Lives in an old peoples home (finding)
        '161832001',  # Weight decreasing (finding)
        '161873000',  # Heavy legs (finding)
        '161874006',  # Heavy feeling (finding)
        '162236007',  # Weakness present (finding)
        '162239000',  # Abdominal weakness (finding)
        '163600007',  # On examination - paresis (weakness) (finding)
        '163686004',  # On examination - gait ataxic (finding)
        '163695007',  # On examination - muscle power reduced (finding)
        '165243005',  # Independent in wheelchair (finding)
        '165244004',  # Minimal help in wheelchair (finding)
        '224960004',  # Tired (finding)
        '225612007',  # Wheelchair bound (finding)
        '238108007',  # Cachexia (finding)
        '248269005',  # Tired on least exertion (finding)
        '248278004',  # Attacks of weakness (finding)
        '248279007',  # Frailty (finding)
        '249888000',  # Weakness of sternomastoid (finding)
        '249937002',  # Truncal muscle weakness (finding)
        '249938007',  # Weakness of back (finding)
        '249939004',  # Proximal muscle weakness (finding)
        '249940002',  # Shoulder girdle weakness (finding)
        '249941003',  # Pelvic girdle weakness (finding)
        '249942005',  # Distal muscle weakness (finding)
        '249943000',  # Weakness of distal arms and legs (finding)
        '249946008',  # Pyramidal type muscle weakness (finding)
        '250002000',  # Rapid fatigue of gait (finding)
        '250003005',  # Low level sensorimotor gait disorder (finding)
        '250015009',  # Arthritic gait (finding)
        '250029005',  # Sensory ataxic gait (finding)
        '250032008',  # Vestibular ataxic gait (finding)
        '250033003',  # Visual ataxic gait (finding)
        '250034009',  # Middle level sensorimotor gait disorder (finding)
        '250038007',  # Retropulsion when walking (finding)
        '250043000',  # High level sensorimotor gait disorder (finding)
        '250044006',  # Cautious gait (finding)
        '250045007',  # Tottering gait (finding)
        '250048009',  # Frontal ataxia (finding)
        '250052009',  # Petren's gait (finding)
        '250054005',  # Frontal gait disorder (finding)
        '262285001',  # Weight decreased (finding)
        '267024001',  # Abnormal weight loss (finding)
        '267031002',  # Tiredness symptom (finding)
        '267032009',  # Tired all the time (finding)
        '268964003',  # On examination - festination-Parkinson gait (finding)
        '271795006',  # Malaise and fatigue (finding)
        '271875007',  # Senile debility (finding)
        '272036004',  # Complaining of debility and malaise (finding)
        '272060000',  # Fatigue - symptom (finding)
        '272062008',  # Complaining of "tired all the time" (finding)
        '275313006',  # Dragging leg (finding)
        '284529003',  # Cardiac cachexia (finding)
        '298283006',  # Hand muscle weakness (finding)
        '300948004',  # Quadriceps weakness (finding)
        '309249007',  # Calf muscle weakness (finding)
        '309257005',  # Excessive weight loss (finding)
        '312444006',  # Spastic paraparesis (finding)
        '314109004',  # Feeling tired (finding)
        '365884000',  # Cerebellar ataxic gait (finding)
        '367391008',  # Malaise (finding)
        '371028005',  # Spastic paresis (finding)
        '373931001',  # Sensation of heaviness in limbs (finding)
        '394616008',  # Unsteady gait (finding)
        '397776000',  # Festinating gait (finding)
        '404904002',  # Frail elderly (finding)
        '413121008',  # Dependent on helper pushing wheelchair (finding)
        '418073009',  # Pseudoparalysis due to generalized arthritis (finding)
        '422868009',  # Unexplained weight loss (finding)
        '426977000',  # Recent weight loss (finding)
        '428116008',  # Multifactorial gait problem (finding)
        '428264009',  # Painful gait (finding)
        '429091008',  # Dependence on biphasic positive airway pressure ventilation (finding)
        '429487005',  # Dependence on continuous positive airway pressure ventilation (finding)
        '431524008',  # Abnormal gait due to impairment of balance (finding)
        '432559006',  # Abnormal gait due to muscle weakness (finding)
        '442099003',  # Psychogenic fatigue (finding)
        '444042007',  # Excessive postexertional fatigue (finding)
        '444932008',  # Dependence on ventilator (finding)
        '448765001',  # Unintentional weight loss (finding)
        '713512009',  # Muscle weakness of upper limb (finding)
        '713514005',  # Muscle weakness of limb (finding)
        '713568000',  # Occasionally tired (finding)
        '713655003',  # Dependence on non-invasive ventilation (finding)
        '784317004',  # Fatigue due to chemotherapy (finding)
        '784318009',  # Fatigue due to radiation therapy (finding)
        '788876001',  # Cachexia due to malignant neoplastic disease (finding)
        '788900007',  # Dependence on artificial heart (finding)
        '931000119107',  # Dependence on supplemental oxygen (finding)
        '60631000119109',  # Dependence on home ventilator (finding)
        '60651000119103',  # Dependence on continuous supplemental oxygen (finding)
        '69161000119103',  # Functional gait abnormality (finding)
        '79021000119104',  # Dependence on aspirator (finding)
        '79031000119101',  # Dependence on respirator (finding)
        '85711000119103',  # Stumbling due to lack of coordination (finding)
        '89201000119106',  # Dependence on supplemental oxygen when ambulating (finding)
        '152921000119101',  # Dependence on respiratory device (finding)
        '250991000119100',  # Stumbling gait (finding)
        '459821000124104',  # McArdle sign (finding)
        '15634971000119107',  # Weakness of bilateral upper limbs (finding)
        '16018391000119104',  # Paresis of left lower limb (finding)
        '16018431000119109',  # Paresis of right lower limb (finding)
        '16419651000119103',  # Dependence on biphasic positive airway pressure ventilation due to central sleep apnea syndrome (finding)
    }




from ..value_set import ValueSet


class CardiacPacer(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent a cardiac pacer device.

    **Data Element Scope:** This value set may use Quality Data Model (QDM) category related to Device.

    **Inclusion Criteria:** Includes only relevant concepts associated with a cardiac pacer device.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1193'
    VALUE_SET_NAME = 'Cardiac Pacer'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '14106009',
        '360127006',
        '360128001',
        '424921004',
        '56961003'
    }

class FrailtyDevice(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent durable medical equipment or devices used by frail patients.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Device.

    **Inclusion Criteria:** Includes only relevant concepts associated with patients that use durable medical equipment or devices for frailty. This is a grouping of HCPCS codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.118.12.1300'
    VALUE_SET_NAME = 'Frailty Device'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '105501005',
        '105503008',
        '105504002',
        '152921000119101',
        '160683008',
        '160684002',
        '160685001',
        '16419651000119103',
        '165243005',
        '165244004',
        '225612007',
        '413121008',
        '429091008',
        '429487005',
        '444932008',
        '60631000119109',
        '60651000119103',
        '713655003',
        '79021000119104',
        '79031000119101',
        '89201000119106',
        '931000119107'
    }


from typing import Any, List, Union

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.value_set.v2020 import Diabetes


class Ccp004v1(ClinicalQualityMeasure):

    class Meta:
        title = 'Diagnosis Of Diabetes'

        version = '2020-04-02v1'

        description = 'All patients with Diagnosis Of Diabetes.'
        information = 'https://canvas-medical.zendesk.com/hc/en-us'

        identifiers = ['CCP004v1']

        types = ['CCP']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_CONDITION,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        show_in_chart = False

        references = ['Canvas Medical CCP, https://canvas-medical.zendesk.com/hc/en-us']

    @cached_property
    def date_of_diagnosis(self) -> Union[Any, str]:
        actives: List = []
        for item in self.patient.conditions.find(Diabetes):
            actives.extend([period['from'] for period in item['periods'] if period['to'] is None])
        return min(actives) if actives else ''

    def in_initial_population(self) -> bool:
        """
        All patients
        """
        return True

    def in_denominator(self) -> bool:
        """
        Patients in the initial population
        """
        return self.in_initial_population()

    def in_numerator(self) -> bool:
        """
        Patients that have been diagnosed with diabetes
        """
        return bool(self.date_of_diagnosis)

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            if self.in_numerator():
                result.due_in = 0
                result.status = STATUS_DUE
                date = arrow.get(self.date_of_diagnosis).format('ddd, MMM Do YYYY')
                narrative = f'{self.patient.first_name} has been diagnosed of diabetes on {date}.'
                result.add_narrative(narrative)
                result.add_recommendation(
                    Recommendation(
                        key='CCP004v1_RECOMMEND_CONTACT',
                        rank=1,
                        button='Communication',
                        title='Contact the patient',
                        narrative=narrative,
                        command={'key': 'schedule'}))
            else:
                result.due_in = -1
                result.status = STATUS_SATISFIED
                narrative = f'{self.patient.first_name} has not been diagnosed of diabetes.'
                result.add_narrative(narrative)

        return result



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AnnualWellnessVisit, AtherosclerosisAndPeripheralArterialDisease, Breastfeeding,
    CerebrovascularDiseaseStrokeTia, Diabetes, EndStageRenalDisease, Ethnicity, HepatitisA,
    HepatitisB_269, HighIntensityStatinTherapy, HospiceCareAmbulatory_1584, Hypercholesterolemia,
    IschemicHeartDiseaseOrOtherRelatedDiagnoses, LdlCholesterol, LiverDisease,
    LowIntensityStatinTherapy, ModerateIntensityStatinTherapy, MyocardialInfarction, OfficeVisit,
    OncAdministrativeSex, OutpatientConsultation, OutpatientEncountersForPreventiveCare,
    PalliativeOrHospiceCare, Payer, PregnancyOrOtherRelatedDiagnoses,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesIndividualCounseling, PreventiveCareServicesInitialOfficeVisit18AndUp,
    PreventiveCareServicesOther, Race, Rhabdomyolysis, StableAndUnstableAngina, StatinAllergen)


class ClinicalQualityMeasure347v4(ClinicalQualityMeasure):
    """
    Statin Therapy for the Prevention and Treatment of Cardiovascular Disease

    Description: Percentage of the following patients - all considered at high risk of
    cardiovascular events - who were prescribed or were on statin therapy during the measurement
    period:
    *Adults aged >= 21 years who were previously diagnosed with or currently have an active
    diagnosis of clinical atherosclerotic cardiovascular disease (ASCVD); OR
    *Adults aged >= 21 years who have ever had a fasting or direct low-density lipoprotein
    cholesterol (LDL-C) level >= 190 mg/dL or were previously diagnosed with or currently have an
    active diagnosis of familial or pure hypercholesterolemia; OR
    *Adults aged 40-75 years with a diagnosis of diabetes with a fasting or direct LDL-C level of
    70-189 mg/dL

    Definition: Clinical atherosclerotic cardiovascular disease (ASCVD) includes:
    *  Acute coronary syndromes
    *  History of myocardial infarction
    *  Stable or unstable angina
    *  Coronary or other arterial revascularization
    *  Stroke or transient ischemic attack (TIA)
    *  Peripheral arterial disease of atherosclerotic origin

    Lipoprotein density cholesterol (LDL-C) result:
    *  A fasting or direct LDL-C laboratory test performed and test result documented in the
    medical record.

    Statin therapy:
    *  Administration of one or more of a group of medications that are used to lower plasma
    lipoprotein levels in the treatment of hyperlipoproteinemia.

    Statin Medication Therapy List (NOTE: List does NOT include dosage):

    [Generic name] (Brand or trade name) and (-) Medication type, if applicable:
    [Atorvastatin] (Lipitor) - Statin
    [Fluvastatin] (Lescol XL or Lescol) - Statin
    [Lovastatin (Mevinolin)](Mevacor or Altoprev) -Statin
    [Pitavastatin]Livalo
    [Pravastatin Sodium] (Pravachol) - Statin
    [Rosuvastatin Calcium] (Crestor) - Statin
    [Simvastatin] (Zocor) - Statin
    [Amlodipine Besylate/Atorvastatin Calcium] (Caduet) - Combination
    [Ezetimibe/Simvastatin] (Vytorin) - Combination

    Some patients may not be appropriate to prescribe or use statin therapy (see
    exceptions and exclusions for a complete list).

    "Statin intolerance is the inability to tolerate a dose of statin required to reduce a person's
    CV risk sufficiently from their baseline risk and could result from different statin related
    side effects including: muscle symptoms, headache, sleep disorders, dyspepsia, nausea, rash,
    alopecia, erectile dysfunction, gynecomastia, and/or arthritis" (Banach et al., 2015, p.2 ).

    Patients that experience symptoms such as these may prefer not to take or continue statin
    therapy and therefore may be exempt from the denominator.

    Rationale: "Cardiovascular disease (CVD) is the leading cause of death in the United States,
    causing approximately 1 of every 3 deaths in the United States in 2015. In 2015, stroke caused
    approximately 1 of every 19 deaths in the United States and the estimated annual costs for CVD
    and stroke were $329.7 billion, including $199.2 billion in direct costs (hospital services,
    physicians and other professionals, prescribed medications, home health care, and other medical
    durables) and $130.5 billion in indirect costs from lost future productivity (cardiovascular
    and stroke premature deaths). CVD costs more than any other diagnostic group" (Benjamin et al.,
    2018).

    Data collected between 2011 and 2014 indicates that more than 94.6 million U.S. adults, 20
    years or older, had total cholesterol levels equal to 200 mg/dL or more, while almost 28.5
    million had levels 240 mg/dL or more (Benjamin et al., 2018). Elevated blood cholesterol is a
    major risk factor for CVD and statin therapy has been associated with a reduced risk of CVD.
    Numerous randomized trials have demonstrated that treatment with a statin reduces LDL-C, and
    reduces the risk of major cardiovascular events by approximately 20 percent (Ference, 2015).

    In 2013, guidelines on the treatment of blood cholesterol to reduce atherosclerotic
    cardiovascular risk in adults were published (see Stone et al., 2014). This guideline was
    published by an Expert Panel, which synthesized evidence from randomized controlled trials to
    identify people most likely to benefit from cholesterol-lowering therapy. The American College
    of Cardiology (ACC)/American Heart Association (AHA) Guideline recommendations are intended to
    provide a strong evidence-based foundation for the treatment of blood cholesterol for the
    primary and secondary prevention and treatment of Atherosclerotic Cardiovascular Disease
    (ASCVD) in adult men and women (21 years of age or older). The document concludes the addition
    of statin therapy reduces the risk of ASCVD among high-risk individuals, defined as follows:
    individuals with clinical ASCVD, with LDL-C >= 190 mg/dL, or with diabetes and LDL-C 70-189
    mg/dL (Stone et al., 2014).

    One study that surveyed U.S. cardiology, primary care, and endocrinology practices found that 1
    in 4 guideline-eligible patients were not on a statin and less than half were on the
    recommended statin intensity. Untreated and undertreated patients had significantly higher
    LDL-C levels than those receiving guideline-directed statin treatment (Navar et al., 2017). The
    Statin Safety Expert Panel that participated in an NLA Statin Safety Task Force meeting in
    October 2013 reaffirms the general safety of statin therapy.

    However, 1 in 10 people who try taking a statin will report some kind of intolerance, most
    commonly muscle aches. Other known low risk circumstances of statin intolerance include side
    effects such as myopathy, cognitive dysfunction, increased hepatic transaminase levels, and new
    onset diabetes. Statin intolerance usually does not involve substantial risk for mortality or
    permanent disability (Guyton et al., 2014). Ultimately, the panel members concluded that for
    most patients requiring statin therapy, the potential benefits of statin therapy outweigh the
    potential risks. In general terms, the benefits of statins to prevent non-fatal myocardial
    infarction, revascularization, stroke, and CVD mortality, far outweighs any potential harm
    related to the drug (Jacobson, 2014).

    Guidance: Numerator instructions and guidance:
    -Current statin therapy use must be documented in the patient's current medication list or
    ordered during the measurement period.
    -ONLY statin therapy meets the measure Numerator criteria (NOT other cholesterol lowering
    medications).
    -Prescription or order does NOT need to be linked to an encounter or visit; it may be called to
    the pharmacy.
    -Statin medication "samples" provided to patients can be documented as "current statin therapy"
    if documented in the medication list in health/medical record.
    -Patients who meet the denominator criteria for inclusion, but are not prescribed or using
    statin therapy, will NOT meet performance for this measure. There is only one performance rate
    calculated for this measure; the weighted average of the three populations.
    -Adherence to statin therapy is not calculated in this measure.

    Denominator Guidance:
    The denominator covers three distinct populations. Use the following process to prevent
    counting patients more than once.

    Denominator Population 1:
    Patients aged >= 21 years at the beginning of the measurement period with clinical ASCVD

    -If YES, meets Denominator Population 1 risk category
    -If NO, screen for next risk category

    Denominator Population 2:
    Patients aged >= 21 years at the beginning of the measurement period who have ever had a
    fasting or direct laboratory test result of LDL-C >= 190 mg/dL or were previously diagnosed
    with or currently have an active diagnosis of familial or pure hypercholesterolemia

    -If YES, meets Denominator Population 2 risk category
    -If NO, screen for next risk category

    Denominator Population 3:
    Patients aged 40 to 75 years at the beginning of the measurement period with Type 1 or Type 2
    diabetes and with a LDL-C result of 70 -189 mg/dL recorded as the highest fasting or direct
    laboratory test result in the measurement year or during the two years prior to the beginning
    of the measurement period

    -If YES, meets Denominator Population 3 risk category
    -If NO, patient does NOT meet Denominator criteria and is NOT eligible for measure inclusion

    Denominator Guidance for Encounter:
    -In order for the patient to be included in the denominator, the patient must have ONE
    denominator-eligible visit, defined as follows:

    --Outpatient encounter visit type
    --Encounter, performed: initial or established office visit, face-to-face interaction,
    preventive care services, or annual wellness visit

    LDL-C Laboratory test result options:
    The measure can be reported for all patients with a documented fasting or direct LDL-C level
    recorded as follows:

    To meet Denominator Population 1:
    There is no LDL-C result required.

    To meet Denominator Population 2:
    If a patient has ANY previous fasting or direct laboratory result of LDL-C >= 190 mg/dL, report
    the highest value >= 190 mg/dL.

    To meet Denominator Population 3:
    If a patient has more than one LDL-C result during the measurement period or during the two
    years before the start of the measurement period, report the highest level recorded during
    either time. The Denominator Exception, "Patients with diabetes who have the most recent
    fasting or direct LDL-C laboratory test result < 70 mg/dL and are not taking statin therapy"
    applies only to Denominator Population 3.

    Intensity of statin therapy in primary and secondary prevention:

    The expert panel of the 2013 ACC/AHA Guidelines (Stone et al., 2014) defines recommended
    intensity of statin therapy on the basis of the average expected LDL-C response to specific
    statin and dose. Although intensity of statin therapy is important in managing cholesterol,
    this measure assesses prescription of ANY statin therapy, irrespective of intensity. Assessment
    of appropriate intensity and dosage documentation added too much complexity to allow inclusion
    of statin therapy intensity in the measure at this time.

    Lifestyle modification coaching:
    A healthy lifestyle is important for the prevention of cardiovascular disease. However,
    lifestyle modification monitoring and documentation added too much complexity to allow its
    inclusion in the measure at this time.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms347v4
    """

    title = 'Statin Therapy for the Prevention and Treatment of Cardiovascular Disease'

    identifiers = ['CMS347v4']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Banach, M., Rizzo, M., Toth, P.,Famier M., Davidson M. H., Al-Rasadi, K., …Mikhailidis, D. P. (2015). Statin intolerance: An attempt at a unified definition. Position paper from an International Lipid Expert Panel. Archives of Medical Science, 11(1), 1-23. doi: 10.5114/aoms.2015.49807',
        'Benjamin, E. J., Virani, S. S., Callaway, C. W., Chamberlain, A. M., Chang, A. R., Cheng, S., …Munter, P. (2018). Heart disease and stroke statistics—2018 update: A report from the American Heart Association. Circulation, 137(12), e67-e492. doi.10.1161/CIR.0000000000000558',
        'Ference, B.A. (2015, March 10). Statins and the risk of developing new-onset Type 2 diabetes: Expert analysis. Retrieved from https://www.acc.org/latest-in-cardiology/articles/2015/03/10/08/10/statins-and-the-risk-of-developing-new-onset-type-2-diabetes',
        'Guyton, J. R., Bays, H. E., Grundy, S. M., Jacobson, T. A. (2014). As assessment by the Statin Intolerance Panel: 2014 update. Journal of Clinical Lipidology, 8(3 Suppl.), S79. doi/10.1016/jacl.2014.03.002',
        'Jacobson, T. A. (2014). Executive summary: NLA Task Force on Statin Safety—2014 update. Journal of Clinical Lipidology, 8(3 Suppl.), S1-S4. doi:10.1016/jacl.2014.03.002',
        'Maddox, T. M., Borden, W. B., Tang, F., Virani, S. S., Oetgen, W. J., Mullen, B., …Rumsfeld, J. S. (2014). Implications of the 2013 ACC/AHA cholesterol guidelines for adults in contemporary cardiovascular practice: insights from the NCDR Pinnacle Registry. Journal of American College of Cardiology, 64(21), 2183-2192. doi:10.1016/j.acc.2014.08.041',
        'Navar, M., Wang, T. Y., Li, S.,Robinson, J. G., Goldberg, A. C., Virani, S., …Peterson, E. D. (2017). Lipid management in contemporary community practice: Results from the Provider Assessment of Lipid Management (PALM) Registry. American Heart Journal, 193, 84-92. doi.10.1016/j.ahj.2017.08.005',
        'Stone, N. J., Robinson, J., Lichtenstein, A. H., Bairey Merz, C., Blum, C. B., Eckel, R. H., …Wilson, P. W. (2014). 2013 ACC/AHA guideline on the treatment of blood cholesterol to reduce atherosclerotic cardiovascular risk in adults: A report of the American College of Cardiology/American Heart Association Task Force on Practice Guidelines Circulation, 129(25, Suppl. 2), S1-S45. doi.10.1161/01.cir.0000437738.63853.7a',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 21 years and older at the beginning of the
        measurement period with a patient encounter during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: All patients who meet one or more of the following criteria (considered at
        "high risk" for cardiovascular events, under ACC/AHA guidelines):

        1) Patients aged >= 21 years at the beginning of the measurement period with clinical ASCVD
        diagnosis

        2) Patients aged >= 21 years at the beginning of the measurement period who have ever had a
        fasting or direct laboratory result of LDL-C >=190 mg/dL or were previously diagnosed with
        or currently have an active diagnosis of familial or pure hypercholesterolemia

        3) Patients aged 40 to 75 years at the beginning of the measurement period with Type 1 or
        Type 2 diabetes and with an LDL-C result of 70-189 mg/dL recorded as the highest fasting or
        direct laboratory test result in the measurement year or during the two years prior to the
        beginning of the measurement period

        Exclusions: Patients who have a diagnosis of pregnancy
        Patients who are breastfeeding
        Patients who have a diagnosis of rhabdomyolysis

        Exceptions: Patients with adverse effect, allergy, or intolerance to statin medication
        Patients who are receiving palliative or hospice care
        Patients with active liver disease or hepatic disease or insufficiency
        Patients with end-stage renal disease (ESRD)
        Patients with diabetes who have the most recent fasting or direct LDL-C laboratory test
        result < 70 mg/dL and are not taking statin therapy
        """
        pass

    def in_numerator(self):
        """
        Numerator: Patients who are actively using or who receive an order (prescription) for
        statin therapy at any point during the measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: This electronic clinical quality measure is intended to align with
        the 2013 ACC/AHA Guideline on the Treatment of Blood Cholesterol (Stone et al., 2014),
        which indicates the use of statins as the first line of cholesterol-lowering medication
        therapy to lower the risk of ASCVD among at-risk populations.

        Recommendations for Treatment of Blood Cholesterol to Reduce Atherosclerotic Cardiovascular
        Risk in Adults - Statin Treatment:
        Secondary Prevention:
        1. High-intensity statin therapy should be initiated or continued as first-line therapy in
        women and men <=75 years of age who have clinical ASCVD, unless contraindicated. (Level of
        Evidence A), (Stone et al., 2014)

        2. In individuals with clinical ASCVD in whom high-intensity statin therapy would otherwise
        be used, when high-intensity statin therapy is contraindicated or when characteristics
        predisposing to statin-associated adverse effects are present, moderate-intensity statin
        should be used as the second option, if tolerated. (Level of Evidence A), (Stone et al.,
        2014)

        Primary Prevention in Individuals >= 21 Years of Age With LDL-C >=190 mg/dL:
        2. Adults >=21 years of age with primary LDL-C >=190 mg/dL should be treated with statin
        therapy. (10-year ASCVD risk estimation is not required.) (Level of Evidence B), (Stone et
        al., 2014)

        Primary Prevention in Individuals With Diabetes and LDL-C 70-189 mg/dL:
        1. Moderate-intensity statin therapy should be initiated or continued for adults 40-75
        years of age with diabetes. (Level of Evidence A), (Stone et al., 2014)
        """
        pass



import pytest
from canvas_workflow_kit.patient import Patient
from canvas_workflow_kit.intervention import (
    Intervention, BannerAlertIntervention
)

from .base import SDKBaseTest


class InterventionTest(SDKBaseTest):

    def test___init__(self):
        intervention1 = Intervention(
            title='Do Something', narrative="I'm not sure what, but do something!",
            href="https://www.canvasmedical.com/")
        self.assertTrue(isinstance(intervention1, Intervention))
        self.assertEqual(intervention1.title, 'Do Something')
        self.assertEqual(intervention1.narrative, "I'm not sure what, but do something!")
        self.assertEqual(intervention1.href, "https://www.canvasmedical.com/")

        intervention2 = Intervention(title='other title')
        self.assertEqual(intervention2.title, 'other title')
        self.assertIsNone(intervention2.narrative)
        self.assertEqual(intervention2.href, '')

        intervention3 = Intervention()
        self.assertIsNone(intervention3.title)
        self.assertIsNone(intervention3.narrative)
        self.assertEqual(intervention3.href, '')


class BannerAlertInterventionTest(SDKBaseTest):
    def test___init__error_no_narrative(self):
        with pytest.raises(TypeError) as init_error:
            BannerAlertIntervention(intent='warning', placement=['scheduling_card'])
        self.assertIn(
            "__init__() missing 1 required positional argument: 'narrative'",
            str(init_error.value))

    def test___init__error_no_narrative_no_placement(self):
        with pytest.raises(TypeError) as init_error:
            BannerAlertIntervention(intent='warning')
        self.assertIn(
            "__init__() missing 2 required positional arguments: 'narrative' and 'placement'",
            str(init_error.value))

    def test___init__error_no_intent_no_placement(self):
        with pytest.raises(TypeError) as init_error:
            BannerAlertIntervention(narrative='Call your mom!')
        self.assertIn(
            "__init__() missing 2 required positional arguments: 'placement' and 'intent'",
            str(init_error.value))

    def test___init__error_no_placement(self):
        with pytest.raises(TypeError) as init_error:
            BannerAlertIntervention(narrative='Call your mom!', intent='warning')
        self.assertIn(
            "__init__() missing 1 required positional argument: 'placement'",
            str(init_error.value))

    def test___init__error_no_intent_no_narrative(self):
        with pytest.raises(TypeError) as init_error:
            BannerAlertIntervention(placement=['scheduling_card'])
        self.assertIn(
            "__init__() missing 2 required positional arguments: 'narrative' and 'intent'",
            str(init_error.value))

    def test___init__error_no_intent(self):
        with pytest.raises(TypeError) as init_error:
            BannerAlertIntervention(placement=['scheduling_card'], narrative='Call your mom!')
        self.assertIn(
            "__init__() missing 1 required positional argument: 'intent'",
            str(init_error.value))

    def test___init__(self):
        alert = BannerAlertIntervention(
            narrative='Call your mom!', intent='warning',
            placement=['scheduling_card', 'appointment_card'])
        self.assertTrue(isinstance(alert, BannerAlertIntervention))
        self.assertEqual(alert.narrative, 'Call your mom!')
        self.assertEqual(alert.intent, 'warning')
        self.assertEqual(alert.placement, ['scheduling_card', 'appointment_card'])

    def test_narrative_limit_error_too_long(self):
        with pytest.raises(ValueError) as narrative_error:
            BannerAlertIntervention(
                narrative='abcdefghijklmnopqrstuvwxyz abcdefghijklmnopqrstuvwxyz abcdefghijklmnopqrstuvwxyz abcdefghijklmnopqrstuvwxyz',
                intent='warning', placement=['scheduling'])
        self.assertEqual(
            'Narrative of Banner Alert Intervention must be between 1 and 90 characters',
            str(narrative_error.value))

    def test_narrative_limit_error_too_short(self):
        with pytest.raises(ValueError) as narrative_error:
            BannerAlertIntervention(narrative='', intent='warning', placement=['scheduling'])
        self.assertIn(
            'Narrative of Banner Alert Intervention must be between 1 and 90 characters',
            str(narrative_error.value))

    def test_narrative(self):
        alert1 = BannerAlertIntervention(
            narrative='H', intent='alert', placement=['scheduling_card'])
        self.assertTrue(isinstance(alert1, BannerAlertIntervention))

        alert2 = BannerAlertIntervention(
            narrative='1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 12',
            intent='info', placement=['scheduling_card'])
        self.assertTrue(isinstance(alert2, BannerAlertIntervention))

    def test_placement_type_error_string(self):
        with pytest.raises(TypeError) as placement_error_string:
            BannerAlertIntervention(narrative='Do pushups!', intent='info', placement='chart')
        self.assertIn(
            'Placement of Banner Alert Intervention must be a List of strings. The options are: ',
            str(placement_error_string.value))

    def test_placement_type_error_empty(self):
        with pytest.raises(ValueError) as placement_error_empty:
            BannerAlertIntervention(narrative='Do pushups!', intent='info', placement=[])
        self.assertIn(
            'Placement of Banner Alert Intervention must have at least one value. The options are: ',
            str(placement_error_empty.value))

    def test_placement_type_error_wrong(self):
        with pytest.raises(ValueError) as placement_error_wrong:
            BannerAlertIntervention(
                narrative='Do pushups!', intent='info', placement=['here'])
        self.assertIn(
            '"here" is not an accepted option for Banner Alert Intervention placement. The options are: ',
            str(placement_error_wrong.value))

    def test_placement_type(self):
        alert1 = BannerAlertIntervention(
            narrative='Do pushups!', intent='info',
            placement=['scheduling_card', 'chart', 'profile', 'appointment_card', 'timeline'])
        self.assertEqual(alert1.placement, ['scheduling_card', 'chart',
                         'profile', 'appointment_card', 'timeline'])

    def test_intent_types(self):
        with pytest.raises(ValueError) as intent_error_empty:
            BannerAlertIntervention(narrative='Do pushups!', intent='', placement=['chart'])
        self.assertIn(
            '"" is not an accepted option for Banner Alert Intervention intent. The options are: ',
            str(intent_error_empty.value))

        with pytest.raises(ValueError) as intent_error_wrong:
            BannerAlertIntervention(narrative='Do pushups!',
                                    intent='negative', placement=['chart'])
        self.assertIn(
            '"negative" is not an accepted option for Banner Alert Intervention intent. The options are: ',
            str(intent_error_wrong.value))

        alert1 = BannerAlertIntervention(narrative='Do pushups!',
                                         intent='info', placement=['chart'])
        self.assertEqual(alert1.intent, 'info')

        alert3 = BannerAlertIntervention(narrative='Do pushups!',
                                         intent='warning', placement=['chart'])
        self.assertEqual(alert3.intent, 'warning')

        alert4 = BannerAlertIntervention(narrative='Do pushups!',
                                         intent='alert', placement=['chart'])
        self.assertEqual(alert4.intent, 'alert')



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AcuteAndSubacuteIridocyclitis, Amblyopia, BestCorrectedVisualAcuityExamUsingSnellenChart,
    BurnConfinedToEyeAndAdnexa, CataractCongenital, CataractMatureOrHypermature,
    CataractPosteriorPolar, CataractSecondaryToOcularDisorders, CentralCornealUlcer,
    CertainTypesOfIridocyclitis, ChoroidalDegenerations, ChoroidalDetachment,
    ChoroidalHemorrhageAndRupture, ChronicIridocyclitis, CloudyCornea, CornealEdema,
    CornealOpacityAndOtherDisordersOfCornea, DegenerationOfMaculaAndPosteriorPole,
    DegenerativeDisordersOfGlobe, DiabeticMacularEdema, DiabeticRetinopathy,
    DisordersOfOpticChiasm, DisordersOfVisualCortex,
    DisseminatedChorioretinitisAndDisseminatedRetinochoroiditis, Ethnicity,
    FocalChorioretinitisAndFocalRetinochoroiditis, Glaucoma,
    GlaucomaAssociatedWithCongenitalAnomaliesDystrophiesAndSystemicSyndromes,
    HereditaryChoroidalDystrophies, HereditaryCornealDystrophies, HereditaryRetinalDystrophies,
    HypotonyOfEye, InjuryToOpticNerveAndPathways, MacularScarOfPosteriorPolar, MorgagnianCataract,
    NystagmusAndOtherIrregularEyeMovements, OncAdministrativeSex, OpenWoundOfEyeball, OpticAtrophy,
    OpticNeuritis, OtherAndUnspecifiedFormsOfChorioretinitisAndRetinochoroiditis,
    OtherBackgroundRetinopathyAndRetinalVascularChanges, OtherDisordersOfOpticNerve,
    OtherEndophthalmitis, OtherProliferativeRetinopathy, PathologicMyopia, Payer,
    PosteriorLenticonus, PriorPenetratingKeratoplasty, PurulentEndophthalmitis, Race,
    RetinalDetachmentWithRetinalDefect, RetinalVascularOcclusion, RetrolentalFibroplasias,
    ScleritisAndEpiscleritis, SeparationOfRetinalLayers, TraumaticCataract, Uveitis,
    VascularDisordersOfIrisAndCiliaryBody, VisualFieldDefects)


class ClinicalQualityMeasure133v9(ClinicalQualityMeasure):
    """
    Cataracts: 20/40 or Better Visual Acuity within 90 Days Following Cataract Surgery

    Description: Percentage of cataract surgeries for patients aged 18 and older with a diagnosis
    of uncomplicated cataract and no significant ocular conditions impacting the visual outcome of
    surgery and had best-corrected visual acuity of 20/40 or better (distance or near) achieved in
    the operative eye within 90 days following the cataract surgery

    Definition: None

    Rationale: In the United States, cataracts affect more than 24 million adults over 40 years
    (National Eye Institute, 2019). According to the American Academy of Ophthalmology (2016),
    cataract surgery has a substantial beneficial impact on visual function and on quality of life.

    1. Scientific basis for measuring visual acuity outcomes after cataract surgery
    The only reason to perform cataract surgery (other than for a limited set of medical
    indications) is to improve a patient's vision and associated functioning. The use of a 20/40
    visual acuity threshold is based on several considerations. First, it is the level for
    unrestricted operation of a motor vehicle in the US. Second, it has been consistently used by
    the FDA in its assessment for approval of intraocular lens (IOL) and other vision devices.
    Third, it is the literature standard to denote success in cataract surgery. Fourth, work by
    West et al. in the Salisbury Eye Study suggests that 20/40 is a useful threshold for 50th
    percentile functioning for several vision-related tasks.

    Most patients achieve excellent visual acuity after cataract surgery (20/40 or better). This
    outcome is achieved consistently through careful attention through the accurate measurement of
    axial length and corneal power and the appropriate selection of an IOL power calculation
    formula. As such, it reflects the care and diligence with which the surgery is assessed,
    planned and executed. Failure to achieve this after surgery in eyes without comorbid ocular
    conditions that would impact the success of the surgery would reflect care that should be
    assessed for opportunities for improvement.

    The exclusion of patients with other ocular and systemic conditions known to increase the risk
    of an adverse outcome reflects the findings of the two published prediction rule papers for
    cataract surgery outcomes, by Mangione et al. (1995) and Steinberg et al. (1994). In both
    papers, the presence of comorbid glaucoma and macular degeneration negatively impacted the
    likelihood of successful outcomes of surgery. Further, as noted in the prior indicator,
    exclusion of eyes with ocular conditions that could impact the success of the surgery would NOT
    eliminate the large majority of eyes undergoing surgery while also minimizing the potential
    adverse selection that might otherwise occur relative to those patients with the most complex
    situations who might benefit the most from having surgery to maximize their remaining vision.

    2. Evidence of a gap in care
    Cataract surgery successfully restores vision in the majority of people who have the procedure.

    Data from a study of 368,256 cataract surgeries show that corrected visual acuity (CDVA) of 0.5
    (20/40) or better was achieved in 94.3% and CDVA of 1.0 (20/20) or better was achieved in 61.3%
    of cases (Lundstrom, Barry, Henry, Rosen & Stenevi, 2013).

    Additionally, data from a UK multi-center Cataract National Dataset found a postoperative
    visual acuity of 6/12 (20/40) or better was achieved for 94.7% of eyes with no co-pathologies
    and in 79.9% of eyes with one or more co-pathologies (Jaycock et al., 2009).

    A rate of 85.5-94.7% of patients achieving a 20/40 or better visual acuity in the context of
    approximately 3 million cataract surgeries in the US annually would mean that between 160,000
    to 435,000 individuals would not achieve a 20/40 or better visual acuity which suggests an
    opportunity for improvement.

    Guidance: This eCQM is an episode-based measure, meaning there may be more than one reportable
    event for a given patient during the measurement period. The level of analysis for this measure
    is each cataract surgery during the measurement period, including instances where more than one
    cataract procedure was performed during the measurement period. Every cataract surgery during
    the measurement period should be counted as a measurable denominator event for the measure
    calculation.

    Only procedures performed during January 1 - September 30 of the reporting period will be
    considered for this measure, in order to determine if 20/40 or better visual acuity has been
    achieved within the 90 days following the cataract procedure. Cataract procedures performed
    during October 1 - December 31 are excluded from the initial population.

    The measure, as written, does not specifically require documentation of laterality. Coding
    limitations in particular clinical terminologies do not currently allow for that level of
    specificity (ICD-10-CM includes laterality, but SNOMED-CT does not uniformly include this
    distinction). Therefore, at this time, it is not a requirement of this measure to indicate
    laterality of the diagnoses, findings or procedures. Available coding to capture the data
    elements specified in this measure has been provided. It is assumed that the eligible
    professional or eligible clinician will record laterality in the patient medical record, as
    quality care and clinical documentation should include laterality.

    This measure is to be reported by the clinician performing the cataract surgery procedure.
    Clinicians who provide only preoperative or postoperative management of cataract patients are
    not eligible for this measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms133v9
    """

    title = 'Cataracts: 20/40 or Better Visual Acuity within 90 Days Following Cataract Surgery'

    identifiers = ['CMS133v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'American Academy of Ophthalmology. (2016). Cataract in the adult eye Preferred Practice Pattern. San Francisco, CA: American Academy of Ophthalmology.',
        'Jaycock, P., Johnston, R. L., Taylor, H., Adams, M., Tole, D. M., Galloway, P., … UK EPR user group (2009). The Cataract National Dataset electronic multi-centre audit of 55,567 operations: Updating benchmark standards of care in the United Kingdom and internationally. Eye, 23(1), 38-49. doi:10.1038/sj.eye.6703015',
        'Lundstrom, M., Barry, P., Henry, Y., Rosen, P., & Stenevi, U. (2013). Visual outcome of cataract surgery; Study from the European Registry of Quality Outcomes for Cataract and Refractive Surgery. Journal of Cataract & Refractive Surgery, 39(5), 673-679. doi:10.1016/j.jcrs.2012.11.026',
        'Mangione, C. M., Orav, J., Lawrence, M. G., Phillips, R.S., Seddon, J.M., & Goldman, L. (1995). Prediction of visual function after cataract surgery: A prospectively validated model. Archives of Ophthalmology, 113(10), 1305-1311. doi:10.1001/archopht.1995.01100100093037',
        'National Eye Institute. (2019). Cataract data and statistics. Retrieved from https://nei.nih.gov/eyedata/cataract',
        'Steinberg, E. P., Tielsch, J. M., Schein, O. D., Javitt, J. C., Sharkey, P., Cassard, S. D., … Damiano, A. M. (1994). National study of cataract surgery outcomes: Variation in 4-month postoperative outcomes as reflected in multiple outcome measures. Ophthalmology, 101(6), 1131-1141. doi:10.1016/s0161-6420(94)31210-3',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All cataract surgeries for patients aged 18 years and older who did not
        meet any exclusion criteria
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Cataract surgeries in patients with significant ocular conditions impacting the
        visual outcome of surgery

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Cataract surgeries with best-corrected visual acuity of 20/40 or better
        (distance or near) achieved in the operative eye within 90 days following cataract surgery

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: This is an outcome measure. As such, there is no statement in the
        guideline specific to this measurement topic.
        """
        pass



from .base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe
import arrow


class TimeFrameTest(SDKBaseTest):

    def setUp(self):
        super().setUp()

    def test___init__(self):
        # correct extremities
        start = arrow.get('1999-06-10', 'YYYY-MM-DD')
        end = arrow.get('2109-06-10', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)
        self.assertEqual(start, tested.start)
        self.assertEqual(end, tested.end)
        # incorrect extremities
        end = arrow.get('1999-06-10', 'YYYY-MM-DD')
        start = arrow.get('2109-06-10', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)
        self.assertEqual(start, tested.start)
        self.assertEqual(end, tested.end)

    def test___str__(self):
        # correct extremities
        start = arrow.get('1999-06-10', 'YYYY-MM-DD')
        end = arrow.get('2109-06-10', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)
        result = str(tested)
        self.assertEqual(
            '<Timeframe start=1999-06-10T00:00:00+00:00, end=2109-06-10T00:00:00+00:00>', result)
        # incorrect extremities
        end = arrow.get('1999-06-10', 'YYYY-MM-DD')
        start = arrow.get('2109-06-10', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)
        result = str(tested)
        self.assertEqual(
            '<Timeframe start=2109-06-10T00:00:00+00:00, end=1999-06-10T00:00:00+00:00>', result)

    def test_cycle(self):
        start = arrow.get('1999-06-10', 'YYYY-MM-DD')
        end = arrow.get('1999-06-20', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)
        self.assertEqual(10, tested.duration)

        start = arrow.get('1999-06-10', 'YYYY-MM-DD')
        end = arrow.get('2001-06-20', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)
        self.assertEqual(365 + 366 + 10, tested.duration)

    def test_increased_by(self):
        start = arrow.get('2018-10-31', 'YYYY-MM-DD')
        end = arrow.get('2018-11-01', 'YYYY-MM-DD')
        tested = Timeframe(start=start, end=end)

        result = tested.increased_by(years=-2)
        self.assertEqual('2016-10-31', result.start.format('YYYY-MM-DD'))
        self.assertEqual('2018-11-01', result.end.format('YYYY-MM-DD'))

        result = tested.increased_by(months=-2)
        self.assertEqual('2018-08-31', result.start.format('YYYY-MM-DD'))
        self.assertEqual('2018-11-01', result.end.format('YYYY-MM-DD'))

        result = tested.increased_by(days=-2)
        self.assertEqual('2018-10-29', result.start.format('YYYY-MM-DD'))
        self.assertEqual('2018-11-01', result.end.format('YYYY-MM-DD'))

        result = tested.increased_by(years=2)
        self.assertEqual('2018-10-31', result.start.format('YYYY-MM-DD'))
        self.assertEqual('2020-11-01', result.end.format('YYYY-MM-DD'))

        result = tested.increased_by(months=2)
        self.assertEqual('2018-10-31', result.start.format('YYYY-MM-DD'))
        self.assertEqual('2019-01-01', result.end.format('YYYY-MM-DD'))

        result = tested.increased_by(days=2)
        self.assertEqual('2018-10-31', result.start.format('YYYY-MM-DD'))
        self.assertEqual('2018-11-03', result.end.format('YYYY-MM-DD'))



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AnaphylacticReactionToCommonBakersYeast, AnaphylacticReactionToDtapVaccine,
    AnaphylacticReactionToHepatitisAVaccine, AntiHepatitisAIggAntigenTest,
    AntiHepatitisBVirusSurfaceAb, DisordersOfTheImmuneSystem, DtapVaccine,
    EncephalopathyDueToChildhoodVaccination, EncounterInpatient, Ethnicity, HepatitisA,
    HepatitisAVaccine, HepatitisB, HepatitisBVaccine, HibVaccine3DoseSchedule,
    HibVaccine4DoseSchedule, Hiv, HomeHealthcareServices, HospiceCareAmbulatory,
    InactivatedPolioVaccineIpv, InfluenzaVaccine, Intussusception,
    MalignantNeoplasmOfLymphaticAndHematopoieticTissue, Measles,
    MeaslesAntibodyTestIggAntibodyPresence, MeaslesAntibodyTestIggAntibodyTiter,
    MeaslesMumpsAndRubellaMmrVaccine, Mumps, MumpsAntibodyTestIggAntibodyPresence,
    MumpsAntibodyTestIggAntibodyTiter, OfficeVisit, OncAdministrativeSex, Payer,
    PneumococcalConjugateVaccine, PreventiveCareEstablishedOfficeVisit0To17,
    PreventiveCareServicesInitialOfficeVisit0To17, Race, RotavirusVaccine3DoseSchedule, Rubella,
    RubellaAntibodyTestIggAntibodyPresence, RubellaAntibodyTestIggAntibodyTiter,
    SevereCombinedImmunodeficiency, VaricellaZoster,
    VaricellaZosterAntibodyTestIggAntibodyPresence, VaricellaZosterAntibodyTestIggAntibodyTiter,
    VaricellaZosterVaccineVzv)


class ClinicalQualityMeasure117v9(ClinicalQualityMeasure):
    """
    Childhood Immunization Status

    Description: Percentage of children 2 years of age who had four diphtheria, tetanus and
    acellular pertussis (DTaP); three polio (IPV), one measles, mumps and rubella (MMR); three or
    four H influenza type B (Hib); three hepatitis B (Hep B); one chicken pox (VZV); four
    pneumococcal conjugate (PCV); one hepatitis A (Hep A); two or three rotavirus (RV); and two
    influenza (flu) vaccines by their second birthday

    Definition: Recommended vaccines: Vaccines and the schedule of vaccines as recommended by the
    Advisory Committee on Immunization Practices (ACIP) for children two years of age. The measure
    may differ slightly from the ACIP recommendations because the measure focuses on immunizations
    that are appropriate by age 2. Also, there may be small differences when there are shortages
    for a particular vaccine.

    Rationale: Infants and toddlers are particularly vulnerable to infectious diseases because
    their immune systems have not built up the necessary defenses to fight infection (Centers for
    Disease Control and Prevention (CDC, 2019). Most childhood vaccines are between 90 and 99
    percent effective in preventing diseases (American Academy of Pediatrics, 2013). Vaccination of
    each U.S. birth cohort with the current childhood immunization schedule prevents approximately
    42,000 deaths and 20 million cases of disease and saves nearly $14 billion in direct costs and
    $69 billion in societal costs each year (Zhou et al., 2014).

    Immunizing a child not only protects that child's health but also the health of the community,
    especially for those who are not immunized or are unable to be immunized due to other health
    complications (Centers for Disease Control and Prevention, 2018).

    Guidance: For the MMR, hepatitis B, VZV and hepatitis A vaccines, numerator inclusion criteria
    include: evidence of receipt of the recommended vaccine; documented history of the illness; or,
    a seropositive test result for the antigen. For the DTaP, IPV, Hib, pneumococcal, rotavirus,
    and influenza vaccines, numerator inclusion criteria include only evidence of receipt of the
    recommended vaccine.

    Patients may be included in the numerator for a particular antigen if they had an anaphylactic
    reaction to the vaccine. Patients may be included in the numerator for the DTaP vaccine if they
    have encephalopathy. Patients may be included in the numerator for the IPV vaccine if they have
    had an anaphylactic reaction to streptomycin, polymyxin B, or neomycin. Patients may be
    included in the numerator for the influenza, MMR, or varicella vaccines if they have cancer of
    lymphoreticular or histiocytic tissue, multiple myeloma, leukemia, have had an anaphylactic
    reaction to neomycin, have immunodeficiency, or have HIV. Patients may be included in the
    numerator for the hepatitis B vaccine if they have had an anaphylactic reaction to common
    baker's yeast.

    The measure allows a grace period by measuring compliance with these recommendations between
    birth and age two.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms117v9
    """

    title = 'Childhood Immunization Status'

    identifiers = ['CMS117v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Centers for Disease Control and Prevention (CDC). (2019). Common Questions About Vaccines. Retrieved October 17, 2019, from https://www.cdc.gov/vaccines/parents/FAQs.html',
        'American Academy of Pediatrics. (2013). Why Immunize Your Child. Retrieved October 17, 2019, from HealthyChildren.org website: http://www.healthychildren.org/english/safety-prevention/immunizations/Pages/Why-Immunize-Your-Child.aspx',
        'Zhou, F., Shefer, A., Wenger, J., Messonnier, M., Wang, L. Y., Lopez, A., … Rodewald, L. (2014). Economic evaluation of the routine childhood immunization program in the United States, 2009. Pediatrics, 133(4), 577–585. https://doi.org/10.1542/peds.2013-0698',
        'Centers for Disease Control and Prevention. (2019). Recommended Child and Adolescent Immunization Schedule for ages 18 years or younger. Retrieved October 22, 2019, from https://www.cdc.gov/vaccines/schedules/downloads/child/0-18yrs-child-combined-schedule.pdf',
        'Centers for Disease Control and Prevention. (2018). Why Are Childhood Vaccines So Important? Retrieved October 23, 2019, from https://www.cdc.gov/vaccines/vac-gen/howvpd.htm',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Children who turn 2 years of age during the measurement period and who
        have a visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients whose hospice care overlaps the measurement period

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Children who have evidence showing they received recommended vaccines, had
        documented history of the illness, had a seropositive test result, or had an allergic
        reaction to the vaccine by their second birthday

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Recommended Child and Adolescent Immunization Schedule
        for ages 18 years or younger, United States 2019 (Centers for Disease Control and
        Prevention, 2019)

        Hepatitis B (HepB)
        "(minimum age: birth)
        Birth dose (monovalent HepB vaccine only)
        -- Mother is HBsAg-negative: 1 dose within 24 hours of birth for all medically stable
        infants ?2,000 grams. Infants <2,000 grams: administer 1 dose at chronological age 1 month
        or hospital discharge.
        -- Mother is HBsAg-positive:
        - Administer HepB vaccine and 0.5 mL of hepatitis B immune globulin (HBIG) (at separate
        anatomic sites) within 12 hours of birth, regardless of birth weight. For infants <2,000
        grams, administer 3 additional doses of vaccine (total of 4 doses) beginning at age 1
        month.
        - Test for HBsAg and anti-HBs at age 9–12 months. If HepB series is delayed, test 1–2
        months after final dose.
        -- Mother’s HBsAg status is unknown:
        - Administer HepB vaccine within 12 hours of birth, regardless of birth weight.
        - For infants <2,000 grams, administer 0.5 mL of HBIG in addition to HepB vaccine within 12
        hours of birth. Administer 3 additional doses of vaccine (total of 4 doses) beginning at
        age 1 month.
        - Determine mother’s HBsAg status as soon as possible. If mother is HBsAg-positive,
        administer 0.5 mL of HBIG to infants ?2,000 grams as soon as possible, but no later than 7
        days of age.
        Routine series
        -- 3-dose series at 0, 1–2, 6–18 months (use monovalent HepB
        vaccine for doses administered before age 6 weeks)
        -- Infants who did not receive a birth dose should begin the
        series as soon as feasible (see Table 2).
        -- Administration of 4 doses is permitted when a combination
        vaccine containing HepB is used after the birth dose.
        -- Minimum age for the final (3rd or 4th ) dose: 24 weeks
        -- Minimum intervals: dose 1 to dose 2: 4 weeks / dose 2 to
        dose 3: 8 weeks / dose 1 to dose 3: 16 weeks (when 4 doses
        are administered, substitute "dose 4" for "dose 3" in these
        calculations) "

        Diphtheria, tetanus, acellular pertussis vaccinations (DTap)
        "(minimum age: 6 weeks [4 years for Kinrix or Quadracel])
        Routine vaccination
        -- 5-dose series at 2, 4, 6, 15–18 months, 4–6 years
        - Prospectively: Dose 4 may be given as early as age 12 months if at least 6 months have
        elapsed since dose 3.
        - Retrospectively: A 4th dose that was inadvertently given as early as 12 months may be
        counted if at least 4 months have elapsed since dose 3."

        Haemophilus influenzae type b (Hib)
        "(minimum age: 6 weeks)
        Routine vaccination
        -- ActHIB, Hiberix, or Pentacel: 4-dose series at 2, 4, 6, 12–15 months
        -- PedvaxHIB: 3-dose series at 2, 4, 12–15 months"

        Poliovirus (inactivated) (IPV)
        " (minimum age: 6 weeks)
        Routine vaccination
        -- 4-dose series at ages 2, 4, 6–18 months, 4–6 years; administer the final dose on or
        after the 4th birthday and at least 6 months after the previous dose.
        -- 4 or more doses of IPV can be administered before the 4th birthday when a combination
        vaccine containing IPV is used. However, a dose is still recommended after the 4th birthday
        and at least 6 months after the previous dose."

        Measles, mumps, and rubella (MMR)
        "(minimum age: 12 months for routine vaccination)
        Routine vaccination
        -- 2-dose series at 12–15 months, 4–6 years
         Dose 2 may be administered as early as 4 weeks after dose 1"

        Pneumococcal
        "(minimum age: 6 weeks [PCV13], 2 years [PPSV23])
        Routine vaccination with PCV13
        -- 4-dose series at 2, 4, 6, 12–15 months"

        Varicella (VAR)
        "(minimum age: 12 months)
        Routine vaccination
        -- 2-dose series: 12–15 months, 4–6 years
        -- Dose 2 may be administered as early as 3 months after dose 1 (a dose administered after
        a 4-week interval may be counted)."

        Hepatitis A (HepA)
        "(minimum age: 12 months for routine vaccination)
        Routine vaccination
        -- 2-dose series (Havrix 6–12 months apart or Vaqta 6–18 months apart, minimum interval 6
        months); a series begun before the 2nd birthday should be completed even if the child turns
        2 before the second dose is administered."

        Rotavirus (RV)
        "(minimum age: 6 weeks)
        Routine vaccination
        -- Rotarix: 2-dose series at 2 and 4 months.
        -- RotaTeq: 3-dose series at 2, 4, and 6 months.
        If any dose in the series is either RotaTeq or unknown, default
        to 3-dose series."

        Influenza (inactivated)  influenza vaccine (IIV)
        "(minimum age: 6 months [IIV], 2 years [LAIV  ], 18 years [RIV])
        Routine vaccination
        -- 1 dose any influenza vaccine appropriate for age and health status annually (2 doses
        separated by at least 4 weeks for children 6 months–8 years who did not receive at least 2
        doses of influenza vaccine before July 1, 2018)"
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.hcc003v1_diabetes_mellitus_with_seconday_complication_suspect import (
    DiabetesOtherClassConditionSuspect,
    DiabetesWithoutComplication,
    Hcc003v1
)
from canvas_workflow_kit.patient_recordset import ConditionRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestHcc003v1(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(Hcc003v1.enabled())

    def test_description(self):
        expected = ('All patients with diabetes, uncomplicated AND a '
                    '2ndary condition often associated with diabetes.')
        self.assertEqual(expected, Hcc003v1._meta.description)

    def test_information(self):
        expected = ('https://canvas-medical.zendesk.com/hc/en-us/articles/360057221174-Diabetes-Mellitus-With-Secondary-Complication-Suspect-HCC003v1')
        self.assertEqual(expected, Hcc003v1._meta.information)

    def test_change_types(self):
        result = Hcc003v1._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test__has_active_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')

        # patient has a active condition after the date --> false
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2018-08-24',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_condition(DiabetesWithoutComplication))

        # patient has a active condition before the date --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested._has_active_condition(DiabetesWithoutComplication))

        # patient has a not active condition before the date --> false
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2017-08-22'
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_condition(DiabetesWithoutComplication))

        # patient has a resolved condition before the end date --> false
        patient.active_only = True
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2018-08-20'
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_condition(DiabetesWithoutComplication))

        # patient has a resolved condition after the end date --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2018-08-25'
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested._has_active_condition(DiabetesWithoutComplication))

        # patient has an other active condition before the date --> false
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119X'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_condition(DiabetesWithoutComplication))

    def test__has_active_class_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')

        # patient has a active condition after the date --> false
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L97123'
            }],
            'periods': [{
                'from': '2018-08-24',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_class_condition(DiabetesOtherClassConditionSuspect))

        # patient has a active condition before the date --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L97123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested._has_active_class_condition(DiabetesOtherClassConditionSuspect))

        # patient has a not active condition before the date --> false
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L97123'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2017-08-23'
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_class_condition(DiabetesOtherClassConditionSuspect))

        # patient has a resolved condition within the period --> false
        patient.active_only = True
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L97123'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2018-08-20'
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_class_condition(DiabetesOtherClassConditionSuspect))

        # patient has a resolved condition after the period --> false
        patient.active_only = True
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L97123'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2018-08-25'
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested._has_active_class_condition(DiabetesOtherClassConditionSuspect))

        # patient has an other active condition before the date --> false
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'M97123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested._has_active_class_condition(DiabetesOtherClassConditionSuspect))

    def test_has_diabetes_with_unspecified_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')
        # patient has a active condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_diabetes_with_unspecified_condition)
        # it is NOT memoize
        tested.patient.conditions = ConditionRecordSet([])
        self.assertFalse(tested.has_diabetes_with_unspecified_condition)

        # patient has NO active condition in HCC
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119X'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertFalse(tested.has_diabetes_with_unspecified_condition)

    def test_has_suspect_eye_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')
        # patient has a active condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'H28'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_suspect_eye_condition)
        # it is memoize
        tested.patient.conditions = ConditionRecordSet([])
        self.assertTrue(tested.has_suspect_eye_condition)

        # patient has NO active condition in HCC
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'H28X'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertFalse(tested.has_suspect_eye_condition)

        # patient has a active class condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'H35123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_suspect_eye_condition)

    def test_has_suspect_neurologic_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')
        # patient has a active condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'G63'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_suspect_neurologic_condition)
        # it is memoize
        tested.patient.conditions = ConditionRecordSet([])
        self.assertTrue(tested.has_suspect_neurologic_condition)

        # patient has NO active condition in HCC
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'G63X'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertFalse(tested.has_suspect_neurologic_condition)

    def test_has_suspect_renal_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')
        # patient has a active condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'N184'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_suspect_renal_condition)
        # it is memoize
        tested.patient.conditions = ConditionRecordSet([])
        self.assertTrue(tested.has_suspect_renal_condition)

        # patient has NO active condition in HCC
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'N184X'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertFalse(tested.has_suspect_renal_condition)

    def test_has_suspect_circulatory_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')
        # patient has a active condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I70123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_suspect_circulatory_condition)
        # it is memoize
        tested.patient.conditions = ConditionRecordSet([])
        self.assertTrue(tested.has_suspect_circulatory_condition)

        # patient has NO active condition in HCC
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I72123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertFalse(tested.has_suspect_circulatory_condition)

    def test_has_suspect_other_condition(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc003v1_diagnosed')
        # patient has a active condition in HCC --> true
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L970123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertTrue(tested.has_suspect_other_condition)
        # it is memoize
        tested.patient.conditions = ConditionRecordSet([])
        self.assertTrue(tested.has_suspect_other_condition)

        # patient has NO active condition in HCC
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L880123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)

        self.assertFalse(tested.has_suspect_other_condition)

    def test_in_initial_population(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # any patient --> true
        patient = self.load_patient('hcc003v1_diagnosed')
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_initial_population())

    def test_in_denominator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has diabetes with unspecified conditions
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        # patient has no diabetes with unspecified conditions
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'E119X'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_numerator(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc003v1_diagnosed')

        # patient has a active eye condition
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'H28'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has a active neurologic condition
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'G63'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has a active renal condition
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'N184'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has a active circulatory condition
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I70123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has a active other complication condition
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'M14123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        # patient has a active other condition
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'XXXX'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

    def test_compute_results(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        # patient has a No active condition diabetes with unspecified complications only
        # --> n/a
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'H28'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

        # patient has a active condition diabetes with unspecified complications only
        # --> satisfied
        patient = self.load_patient('hcc003v1_diagnosed')
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        self.assertEqual('', result.narrative)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)

        # patient has a active condition diabetes with unspecified complications + eye suspect
        # --> due
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = patient.conditions + ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'H28'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has Diabetes without complications AND '
                f'an eye condition commonly caused by diabetes on the Conditions list.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': (f'Consider updating the Diabetes without complications (E11.9) '
                          f'to Diabetes with secondary eye disease as clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_EYE',
                'rank': 1,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has a active condition diabetes with
        # unspecified complications + neurologic suspect --> due
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = patient.conditions + ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'G63'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has Diabetes without complications AND '
                f'a neurological condition commonly caused by diabetes on the Conditions list.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': ('Consider updating the Diabetes without '
                          'complications (E11.9) to Diabetes with '
                          'secondary neurological sequela as '
                          'clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_NEUROLOGICAL_SEQUELA',
                'rank': 2,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has a active condition diabetes with unspecified complications + renal suspect
        # --> due
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = patient.conditions + ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'N184'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has Diabetes without complications AND '
                f'a chronic renal condition commonly caused by diabetes on the Conditions list.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': (f'Consider updating the Diabetes without complications (E11.9) '
                          f'to Diabetes with secondary renal disease as clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_RENAL_DISEASE',
                'rank': 3,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has a active condition diabetes with
        # unspecified complications + circulatory suspect --> due
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = patient.conditions + ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I79123'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has Diabetes without complications AND '
                f'a circulatory condition commonly caused by diabetes on the Conditions list.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': ('Consider updating the Diabetes without '
                          'complications (E11.9) to Diabetes with '
                          'secondary circulatory disorder as '
                          'clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_CIRCULATORY_DISORDER',
                'rank': 4,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has a active condition diabetes with
        # unspecified complications + other suspect --> due
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = patient.conditions + ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'L98456'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }]
        }])
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has Diabetes without complications AND '
                f'an another condition commonly caused by diabetes on the Conditions list.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': ('Consider updating the Diabetes without '
                          'complications (E11.9) to Diabetes with '
                          'other secondary complication as '
                          'clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_COMPLICATION',
                'rank': 5,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

        # patient has a active condition diabetes with
        # unspecified complications + eye suspect + renal suspect --> due
        patient = self.load_patient('hcc003v1_diagnosed')
        patient.conditions = patient.conditions + ConditionRecordSet([
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'system': 'ICD-10',
                    'code': 'N184'
                }],
                'periods': [{'from': '2018-08-23', 'to': None}]
            },
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'system': 'ICD-10',
                    'code': 'H28'
                }],
                'periods': [{'from': '2018-08-23', 'to': None}]
            }
        ])  # yapf: disable
        tested = Hcc003v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (f'Jim has Diabetes without complications AND '
                f'an eye condition commonly caused by diabetes on the Conditions list.\n'
                f'Jim has Diabetes without complications AND '
                f'a chronic renal condition commonly caused by diabetes on the Conditions list.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': (f'Consider updating the Diabetes without complications (E11.9) '
                          f'to Diabetes with secondary eye disease as clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_EYE',
                'rank': 1,
                'button': 'Diagnose',
            },
            {
                'title': (f'Consider updating the Diabetes without complications (E11.9) '
                          f'to Diabetes with secondary renal disease as clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC003v1_RECOMMEND_DIAGNOSE_RENAL_DISEASE',
                'rank': 3,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

    def helper_exam_recommended(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['key'], recommendation.command['key'])



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AllergyToBetaBlockerTherapy, Arrhythmia, Asthma, AtrioventricularBlock,
    BetaBlockerTherapyForLvsd, BetaBlockerTherapyIngredient, Bradycardia, CardiacPacerInSitu,
    CareServicesInLongTermResidentialFacility, DischargeServicesHospitalInpatient,
    EjectionFraction, Ethnicity, HeartFailure, HeartRate, HomeHealthcareServices, Hypotension,
    IntoleranceToBetaBlockerTherapy, LeftVentricularSystolicDysfunction, ModerateOrSevereLvsd,
    NursingFacilityVisit, OfficeVisit, OncAdministrativeSex, OutpatientConsultation,
    PatientProviderInteraction, Payer, Race)


class ClinicalQualityMeasure144v9(ClinicalQualityMeasure):
    """
    Heart Failure (HF): Beta-Blocker Therapy for Left Ventricular Systolic Dysfunction (LVSD)

    Description: Percentage of patients aged 18 years and older with a diagnosis of heart failure
    (HF) with a current or prior left ventricular ejection fraction (LVEF) < 40% who were
    prescribed beta-blocker therapy either within a 12-month period when seen in the outpatient
    setting OR at each hospital discharge

    Definition: Prescribed-Outpatient setting: prescription given to the patient for beta-blocker
    therapy at one or more visits in the measurement period OR patient already taking beta-blocker
    therapy as documented in current medication list.

    Prescribed-Inpatient setting: prescription given to the patient for beta-blocker therapy at
    discharge OR beta-blocker therapy to be continued after discharge as documented in the
    discharge medication list.

    LVEF < 40% corresponds to qualitative documentation of moderate dysfunction or severe
    dysfunction.

    Rationale: Beta-blockers are recommended for all patients with stable heart failure and left
    ventricular systolic dysfunction, unless contraindicated. Treatment should be initiated as soon
    as a patient is diagnosed with left ventricular systolic dysfunction and does not have low
    blood pressure, fluid overload, or recent treatment with an intravenous positive inotropic
    agent. Beta-blockers have been shown to lessen the symptoms of heart failure, improve the
    clinical status of patients, reduce future clinical deterioration, and decrease the risk of
    mortality and the combined risk of mortality and hospitalization.

    Guidance: This eCQM is to be reported as patient-based or episode-based, depending on the
    clinical setting. To satisfy this measure, it must be reported for all heart failure patients
    at least once during the measurement period if seen in the outpatient setting. If the patient
    has an eligible inpatient discharge during the measurement period, as defined in the measure
    logic, it is expected to be reported at each hospital discharge.

    A range value should satisfy the logic requirement for 'Ejection Fraction' as long as the
    ranged observation value clearly meets the less than 40% threshold noted in the denominator
    logic. A range that is inclusive of or greater than 40% would not meet the measure requirement.

    Beta-blocker therapy:
    -For patients with prior LVEF < 40%, beta-blocker therapy should include bisoprolol,
    carvedilol, or sustained release metoprolol succinate.

    The requirement of two or more visits used in Population Criteria 1 is to establish that the
    eligible professional or eligible clinician has an existing relationship with the patient.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms144v9
    """

    title = 'Heart Failure (HF): Beta-Blocker Therapy for Left Ventricular Systolic Dysfunction (LVSD)'

    identifiers = ['CMS144v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Yancy, C. W., Jessup, M., Bozkurt, B., Butler, J., Casey, D. E., Drazner, M. H., … Wilkoff, B. L. (2013). 2013 ACCF/AHA Guideline for the Management of Heart Failure A Report of the American College of Cardiology Foundation/American Heart Association Task Force on Practice Guidelines. Circulation, 128(16), e240–e327. doi: https://doi.org/10.1161/CIR.0b013e31829e8776',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 18 years and older with a diagnosis of heart failure
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population with a current or prior LVEF < 40%

        Exclusions: None

        Exceptions: Documentation of medical reason(s) for not prescribing beta-blocker therapy
        (e.g., low blood pressure, fluid overload, asthma, patients recently treated with an
        intravenous positive inotropic agent, allergy, intolerance, other medical reasons).

        Documentation of patient reason(s) for not prescribing beta-blocker therapy (e.g., patient
        declined, other patient reasons).

        Documentation of system reason(s) for not prescribing beta-blocker therapy (e.g., other
        reasons attributable to the healthcare system).
        """
        pass

    def in_numerator(self):
        """
        Numerator: Patients who were prescribed beta-blocker therapy either within a 12-month
        period when seen in the outpatient setting OR at each hospital discharge

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Use of 1 of the 3 beta blockers proven to reduce mortality (e.g.,
        bisoprolol, carvedilol, and sustained-release metoprolol succinate) is recommended for all
        patients with current or prior symptoms of HFrEF [heart failure with reduced ejection
        fraction], unless contraindicated, to reduce morbidity and mortality (Class I, Level of
        Evidence: A) (ACCF/AHA, 2013).

        Treatment with a beta blocker should be initiated at very low doses [see excerpt from
        guideline table below] followed by gradual increments in dose if lower doses have been well
        tolerated... Clinicians should make every effort to achieve the target doses of the beta
        blockers shown to be effective in major clinical trials. Even if symptoms do not improve,
        long-term treatment should be maintained to reduce the risk of major clinical events.
        Abrupt withdrawal of treatment with a beta blocker can lead to clinical deterioration and
        should be avoided (ACCF/AHA, 2013).

        Drugs Commonly Used for Stage C HFrEF (abbreviated to align with focus of measure to
        include only Beta-blocker therapy)
        Drug                               Initial Daily Dose(s)       Maximum Dose(s)        Mean
        Doses Achieved in Clinical Trials

        Beta Blockers
        Bisoprolol                         1.25 mg once               10 mg once
        8.6 mg/d
        Carvedilol                         3.125 mg twice            50 mg twice
        37 mg/d
        Carvedilol  CR                  10 mg once                  80 mg once
        N/A
        Metoprolol succinate        12.5 to 25 mg once      200 mg once                159 mg/d
        extended release
        (metoprolol CR/XL)

        For the hospitalized patient:
        In patients with HFrEF experiencing a symptomatic exacerbation of HF requiring
        hospitalization during chronic maintenance treatment with GDMT [guideline-directed medical
        therapy; GDMT represents optimal medical therapy as defined by ACCF/AHA guideline-
        recommended therapies (primarily Class I)], it is recommended that GDMT be continued in the
        absence of hemodynamic instability or contraindications (Class I, Level of Evidence: B)
        (ACCF/AHA, 2013).

        Initiation of beta-blocker therapy is recommended after optimization of volume status and
        successful discontinuation of intravenous diuretics, vasodilators, and inotropic agents.
        Beta-blocker therapy should be initiated at a low dose and only in stable patients. Caution
        should be used when initiating beta blockers in patients who have required inotropes during
        their hospital course (Class I, Level of Evidence: B) (ACCF/AHA, 2013).
        """
        pass



from ..value_set import ValueSet


class AverageNumberOfDrinksPerDrinkingDay(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent a patient's number of alcoholic drinks per drinking day.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying a quantitative observation of the reported number of alcoholic drinks per drinking day. This is a grouping of a LOINC code.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.106.12.1018'
    VALUE_SET_NAME = 'Average Number of Drinks per Drinking Day'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '11287-0'
    }

class FallsScreening(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent assessment tools or instruments used to quantify fall risk.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with fall risk assessment tools and screening assessments used to evaluate fall risk such as the Get Up and Go Test, Morse falls risk assessment, multidisciplinary team falls assessment, and history of falls. This is a grouping of LOINC and SNOMED CT codes.

    **Exclusion Criteria:** Excludes codes that indicate falls for children or falls that take place in an inpatient setting.
    """

    OID = '2.16.840.1.113883.3.464.1003.118.12.1028'
    VALUE_SET_NAME = 'Falls Screening'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '52552-7',
        '57254-5',
        '59454-9',
        '73830-2'
    }

class HipDysfunctionAndOsteoarthritisOutcomeScoreForJointReplacementHoosjr(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent the Hip Dysfunction and Osteoarthritis Outcome Score for Joint Replacement (HOOS Jr.) total interval score.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with identifying the Hip Dysfunction and Osteoarthritis Outcome Score for Joint Replacement (HOOS Jr.) total interval score. This is a grouping of a LOINC code.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.118.12.1210'
    VALUE_SET_NAME = 'Hip Dysfunction and Osteoarthritis Outcome Score for Joint Replacement [HOOSJR]'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '82323-7'
    }

class HistoryOfHipFractureInParent(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent a family history of fracture in a parent.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with a patient has a family history of fracture in a parent. This is a grouping of SNOMED CT codes.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.113.12.1040'
    VALUE_SET_NAME = 'History of hip fracture in parent'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '391096007',
        '416072008',
        '445121000124105',
        '445501000124107'
    }

class KneeInjuryAndOsteoarthritisOutcomeScoreForJointReplacementKoosjr(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent the Knee Injury and Osteoarthritis Outcome Score for Joint Replacement (KOOS Jr.) total interval score.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with the Knee Injury and Osteoarthritis Outcome Score for Joint Replacement (KOOS Jr.) score. This is a grouping of a LOINC code.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.118.12.1218'
    VALUE_SET_NAME = 'Knee Injury and Osteoarthritis Outcome Score for Joint Replacement [KOOSJR]'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '82324-5'
    }

class Phq9AndPhq9MTools(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent the PHQ 9 and PHQ 9M depression assessment scores for  adults and adolescents.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with LOINC codes that indicate a completed PHQ 9 and PHQ 9M depression assessment tool for adults and adolescents with a summary score.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.67.1.101.1.263'
    VALUE_SET_NAME = 'PHQ 9 and PHQ 9M Tools'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '44261-6',
        '89204-2'
    }

class SexuallyActive(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent vaginal intercourse.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) datatype related to Assessment, Performed.

    **Inclusion Criteria:** Includes only relevant concepts associated with concepts that indicate whether a patient has had vaginal intercourse.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.464.1003.121.12.1040'
    VALUE_SET_NAME = 'Sexually Active'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '64728-9'
    }


import importlib
import inspect
import requests
import json

from typing import Optional, Union

from pathlib import Path

from canvas_workflow_kit.patient import Patient
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.api import fetch_patient
from canvas_workflow_kit.internal.string import camelcase


def parse_class_from_python_source(source):
    """
    Parse the python file. Return a ClinicalQualityMeasure class if only one is
    found in the file.
    """

    spec = importlib.util.spec_from_loader('helper', loader=None)
    helper = importlib.util.module_from_spec(spec)

    exec(source, helper.__dict__)

    clinical_quality_measures = []

    for cls_name, cls in inspect.getmembers(helper, inspect.isclass):
        if cls.__module__ != 'helper':
            continue
        if issubclass(cls, ClinicalQualityMeasure):
            clinical_quality_measures.append(cls)

    if len(clinical_quality_measures) == 0:
        raise SyntaxError("No clinical quality measures found.")
    elif len(clinical_quality_measures) > 1:
        raise SyntaxError("More than one clinical quality measures found.")

    return clinical_quality_measures[0]


def send_notification(url, payload={}, headers={}):
    return requests.post(url, data=payload, headers=headers)


def load_patient(patient_key: str, config_name: Optional[str] = None):
    """
    Load a patient's data off a server to test.
    """
    response_json = fetch_patient(patient_key=patient_key, config_name=config_name)
    return Patient(response_json)


def load_local_patient(patient_path: Union[str, Path]) -> Patient:
    """
    Load a patient's data from a local downloaded directory.
    """
    # cast to Path
    if type(patient_path) is str:
        patient_path = Path(patient_path)

    combined_json = {}
    for filepath in patient_path.glob('*.json'):
        with filepath.open('r') as file:
            combined_json[camelcase(filepath.stem)] = json.load(file)  # type: ignore

    if not combined_json:
        raise FileNotFoundError(f'No JSON files were found in "{patient_path}"')

    return Patient(combined_json)



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    ConsultantReport, Ethnicity, OfficeVisit, OncAdministrativeSex, OphthalmologicalServices,
    Payer, PreventiveCareEstablishedOfficeVisit0To17,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit0To17, PreventiveCareServicesInitialOfficeVisit18AndUp,
    Race, Referral)


class ClinicalQualityMeasure50v9(ClinicalQualityMeasure):
    """
    Closing the Referral Loop: Receipt of Specialist Report

    Description: Percentage of patients with referrals, regardless of age, for which the referring
    provider receives a report from the provider to whom the patient was referred

    Definition: Referral: A request from one physician or other eligible provider to another
    practitioner for evaluation, treatment, or co-management of a patient's condition. This term
    encompasses referral and consultation as defined by Centers for Medicare & Medicaid Services.

    Report: A written document prepared by the eligible clinician (and staff) to whom the patient
    was referred and that accounts for his or her findings, provides summary of care information
    about findings, diagnostics, assessments and/or plans of care, and is provided to the referring
    eligible clinician.

    Rationale: Problems in the outpatient referral and consultation process have been documented,
    including lack of timeliness of information and inadequate provision of information between the
    specialist and the requesting physician (Gandhi et al., 2000; Forrest et al., 2000; Stille,
    Jerant, Bell, Meltzer, & Elmore, 2005). In a study of physician satisfaction with the
    outpatient referral process, Gandhi et al. (2000) found that 68% of specialists reported
    receiving no information from the primary care provider prior to referral visits, and 25% of
    primary care providers had still not received any information from specialists 4 weeks after
    referral visits. In another study of 963 referrals (Forrest et al., 2000), pediatricians
    scheduled appointments with specialists for only 39% and sent patient information to the
    specialists in only 51% of the time.

    In a 2006 report to Congress, the Medicare Payment Advisory Commission (MedPAC) found that care
    coordination programs improved quality of care for patients, reduced hospitalizations, and
    improved adherence to evidence-based care guidelines, especially among patients with diabetes
    and CHD. Associations with cost-savings were less clear; this was attributed to how well the
    intervention group was chosen and defined, as well as the intervention put in place.
    Additionally, cost-savings were usually calculated in the short-term, while some argue that the
    greatest cost-savings accrue over time (MedPAC, 2006).

    Improved mechanisms for information exchange could facilitate communication between providers,
    whether for time-limited referrals or consultations, on-going co-management, or during care
    transitions. For example, a study by Branger, van’t Hooft, van der Wouden, Moorman & van Bemmel
    (1999) found that an electronic communication network that linked the computer-based patient
    records of physicians who had shared care of patients with diabetes significantly increased
    frequency of communications between physicians and availability of important clinical data.
    There was a 3-fold increase in the likelihood that the specialist provided written
    communication of results if the primary care physician scheduled appointments and sent patient
    information to the specialist (Forrest et al., 2000).

    Care coordination is a focal point in the current health care reform and our nation's
    ambulatory health information technology (HIT) framework. The National Priorities Partnership
    (2008) recently highlighted care coordination as one of the most critical areas for development
    of quality measurement and improvement.

    Guidance: The provider who refers the patient to another provider is the provider who should be
    held accountable for the performance of this measure.

    The provider to whom the patient was referred should be the same provider that sends the
    report.

    If there are multiple referrals for a patient during the measurement period, use the first
    referral.

    The consultant report that will fulfill the referral should be completed after the referral and
    should be related to the referral for which it is attributed. If there are multiple consultant
    reports received by the referring provider which pertain to a particular referral, use the
    first consultant report to satisfy the measure. Eligible professionals or eligible clinicians
    reporting on this measure should note that all data for the reporting year is to be submitted
    by the deadline established by CMS. Therefore, eligible professionals or eligible clinicians
    who refer patients towards the end of the reporting period (i.e., November - December), should
    request that providers to whom they referred their patients share their consult reports as soon
    as possible in order for those patients to be counted in the measure numerator during the
    measurement period. When providers to whom patients are referred communicate the consult report
    as soon as possible with the referring providers, it ensures that the communication loop is
    closed in a timely manner and that the data are included in the submission to CMS.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms50v9
    """

    title = 'Closing the Referral Loop: Receipt of Specialist Report'

    identifiers = ['CMS50v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Branger, P. J., van\'t Hooft, A., van der Wouden, J. C., Moorman, P.W., & van Bemmel, J.H. (1999). Shared care for diabetes: Supporting communication between primary and secondary care. International Journal of Medical Informatics, 53(2-3), 133-142. doi: 10.1016/s1386-5056(98)00154-3',
        'Forrest, C. B., Glade, G. B., Baker, A. E., Bocian, A., von Schrader, S., & Starfield, B. (2000). Coordination of specialty referrals and physician satisfaction with referral care. Archives of Pediatrics and Adolescent Medicine, 154(5), 499-506. doi: 10.1001/archpedi.154.5.499',
        'Gandhi, T. K., Sittig, D. F., Franklin, M., Sussman, A.J., Fairchild, D.G., & Bates, D.W. (2000). Communication breakdown in the outpatient referral process. Journal of General Internal Medicine, 15(9), 626-631. doi: 10.1046/j.1525-1497.2000.91119.x',
        'MedPAC. (2006, March). Report to the Congress: Medicare payment policy. Retrieved from http://medpac.gov/docs/default-source/reports/Mar06_EntireReport.pdf?sfvrsn=0',
        'National Priorities Partnership. (2008). National priorities and goals: Aligning our efforts to transform America’s healthcare. Washington, DC: National Quality Forum.',
        'Stille, C. J., Jerant, A., Bell, D., Meltzer, D., & Elmore, J.G. (2005). Coordinating care across diseases, settings, and clinicians: A key role for the generalist in practice. Annals of Internal Medicine, 142(8), 700-708. doi: 10.7326/0003-4819-142-8-200504190-00038',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Number of patients, regardless of age, who were referred by one
        provider to another provider, and who had a visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Number of patients with a referral, for which the referring provider received a
        report from the provider to whom the patient was referred

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: None
        """
        pass



# flake8: noqa
from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.value_set.v2018 import (OfficeVisit, PsychVisit, FaceToFaceInteractionNoEd, Dysthymia,
                                 MajorDepression, CareServicesInLongTermResidentialFacility,
                                 PalliativeCareEncounter, PalliativeCare, BipolarDisorder,
                                 PersonalityDisorder)


class DepressionUtilizationPhq9(ClinicalQualityMeasure):
    """
    Guidance: If a patient has a qualifying diagnosis and encounter in more than one of the 4 month
    periods within the measurement year, the patient must be counted (denominator and numerator) in
    each qualifying 4 month period. For example, a patient could be counted in the first and third
    4 month periods.
    """

    # this protocol was created before an overhaul and needs finishing
    # to work
    @classmethod
    def enabled(cls):
        return False

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    def in_initial_population(self):

        def age_range():
            # TODO is this correct or can the patient turn 18 during the measurement period?
            return self.patient.age_at(self.timeframe.start) >= 18

        def has_office_visit():
            return self.patient.conditions_within(self.timeframe, OfficeVisit)

        def has_psych_visit():
            return self.patient.conditions_within(self.timeframe, PsychVisit)

        def has_face_to_face_no_ed():
            return self.patient.conditions_within(self.timeframe, FaceToFaceInteractionNoEd)

        def has_qualifying_diagnosis():
            # TODO is this correct or does the diagnosis have to be within the timeframe, not just
            # before it ends?
            return any([
                self.patient.conditions_within(self.timeframe, Dysthymia),
                self.patient.conditions_within(self.timeframe, MajorDepression),
            ])

        return all([
            age_range(),
            has_qualifying_diagnosis(),
            any([
                has_office_visit(),
                has_psych_visit(),
                has_face_to_face_no_ed(),
            ]),
        ])

    def in_denominator(self):
        if not self.in_initial_population():
            return False

        def long_term_care_resident():
            return self.patient.conditions_after_one_year_before_timeframe_end(
                self.timeframe, CareServicesInLongTermResidentialFacility)

        def received_palliative_care():
            return any([
                self.patient.conditions_after_one_year_before_timeframe_end(
                    self.timeframe, PalliativeCare),
                self.patient.conditions_after_one_year_before_timeframe_end(
                    self.timeframe, PalliativeCareEncounter),
            ])

        def has_disqualifying_disorder():
            return any([
                self.patient.conditions_before(self.timeframe.end, BipolarDisorder),
                self.patient.conditions_before(self.timeframe.end, PersonalityDisorder),
            ])

        # TODO 'patient characteristic: expired' not in value set
        def deceased():
            return False

        return all([
            not deceased(),
            not received_palliative_care(),
            not long_term_care_resident(),
            not has_disqualifying_disorder(),
        ])

    def in_numerator(self):
        pass
        # return self.patient.has_any_within(timeframe=self.timeframe,
        # **value_sets.phq_tool.values)
